# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sat Dec 13 00:41:15 2014

@author: Vidar Tonaas Fauske
"""

import os
import sys
import glob
import importlib
import warnings
import traceback

from hyperspyui.log import logger
from hyperspyui.plugins.plugin import Plugin
from hyperspyui.settings import Settings
from hyperspyui.util import AttributeDict


class ReadOnlyDict(dict):
    _readonly = False

    def __setitem__(self, key, value):

        if self._readonly:
            raise TypeError("This dictionary is read only")
        return dict.__setitem__(self, key, value)

    def __delitem__(self, key):

        if self._readonly:
            raise TypeError("This dictionary is read only")
        return dict.__delitem__(self, key)

    def pop(self, *args, **kwargs):
        if self._readonly:
            raise TypeError("This dictionary is read only")
        return dict.pop(*args, **kwargs)

    def update(self, *args, **kwargs):
        if self._readonly:
            raise TypeError("This dictionary is read only")
        return dict.update(*args, **kwargs)


class PluginManager(object):

    def __init__(self, main_window):
        """
        Initializates the manager, and performs discovery of plugins
        """
        self.plugins = AttributeDict()
        self.ui = main_window
        self._enabled = {}
        self.settings = Settings(self.ui, group="General")
        self.settings.set_default("extra_plugin_directories", "")
        self.enabled_store = Settings(self.ui, group="PluginManager/enabled")

        self.discover()

    @property
    def enabled(self):
        """Returns a read-only dictionary showing the enabled/disabled state
        of all plugins.
        """
        d = ReadOnlyDict()
        for name, (enabled, _) in d.items():
            d[name] = enabled
        d._readonly = True
        return d

    def enable_plugin(self, name, value=True):
        """Enable/disable plugin functionality. Also loads/unloads plugin. If
        enabling and the plugin is already loaded, this will reload the plugin.
        """
        self.enabled_store[name] = value
        ptype = self._enabled[name][1]
        self._enabled[name] = (value, ptype)
        if name in self.plugins:
            self.unload(self.plugins[name])
        if value:
            self.load(ptype)

    def disable_plugin(self, name):
        """Disable plugin functionality. Also unloads plugin.
        """
        self.enable_plugin(name, False)

    def _import_plugin_from_path(self, name, path):
        try:
            mname = "hyperspyui.plugins." + name
            if sys.version_info >= (3, 5):
                import importlib.util
                spec = importlib.util.spec_from_file_location(mname, path)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
            else:
                from importlib.machinery import SourceFileLoader
                loader = SourceFileLoader(mname, path)
                loader.load_module()
        except Exception:
            self.warn("import", path)

    def discover(self):
        """Auto-discover all plugins defined in plugin directory.
        """
        import hyperspyui.plugins as plugins
        for plug in plugins.__all__:
            try:
                __import__('hyperspyui.plugins.' + plug, globals())

            except Exception:
                self.warn("import", plug)

        # Import any plugins in extra dirs.
        extra_paths = self.settings['extra_plugin_directories']
        if extra_paths:
            extra_paths = extra_paths.split(os.path.pathsep)
            for path in extra_paths:
                if not os.path.isdir(path):
                    path = os.path.dirname(path)
                modules = glob.glob(path + "/*.py")
                # TODO: In release form, we should consider supporting
                # compiled plugins in pyc/pyo format
                # modules.extend(glob.glob(os.path.dirname(__file__)+"/*.py?"))
                # If so, ensure that duplicates are removed (picks py over pyc)
                modules = [m for m in modules
                           if not os.path.basename(m).startswith('_')]

                for m in modules:
                    name = os.path.splitext(os.path.basename(m))[0]
                    self._import_plugin_from_path(name, m)

        master = Plugin
        self.implementors = sorted(self._inheritors(master), key=lambda x: x.name)
        logger.debug("Found plugins: %s", self.implementors)

    @staticmethod
    def _inheritors(klass):
        """Return all defined classes that inherit from 'klass'.
        """
        subclasses = set()
        work = [klass]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return subclasses

    def warn(self, f_name, p_name, category=RuntimeWarning):
        tbf = ''.join(traceback.format_exception(*sys.exc_info())[2:])
        warnings.warn(("Exception in {0} of hyperspyui plugin " +
                       "\"{1}\" error:\n{2}").format(f_name, p_name, tbf),
                      RuntimeWarning, 2)

    def init_plugins(self):
        self.plugins.clear()
        for plug_type in self.implementors:
            try:
                self._load_if_enabled(plug_type)
            except Exception:
                self.warn("initialization", plug_type.name)

    def create_actions(self):
        for p in self.plugins.values():
            try:
                p.create_actions()
            except Exception:
                self.warn(sys._getframe().f_code.co_name, p.name)

    def create_menu(self):
        for p in self.plugins.values():
            try:
                p.create_menu()
            except Exception:
                self.warn(sys._getframe().f_code.co_name, p.name)

    def create_tools(self):
        for p in self.plugins.values():
            try:
                p.create_tools()
            except Exception:
                self.warn(sys._getframe().f_code.co_name, p.name)

    def create_toolbars(self):
        for p in self.plugins.values():
            try:
                p.create_toolbars()
            except Exception:
                self.warn(sys._getframe().f_code.co_name, p.name)

    def create_widgets(self):
        for p in self.plugins.values():
            try:
                p.create_widgets()
            except Exception:
                self.warn(sys._getframe().f_code.co_name, p.name)

    def _load_if_enabled(self, p_type):
        if p_type is None or p_type.name is None:
            return
        if self.enabled_store[p_type.name] is None:
            # Init setting to True on first encounter
            self.enabled_store[p_type.name] = True
        enabled = self.enabled_store[p_type.name, bool]
        self._enabled[p_type.name] = (enabled, p_type)
        if enabled:
            # Init
            logger.debug("Initializing plugin: %s", p_type)
            p = p_type(self.ui)
            self.plugins[p.name] = p
            logger.debug("Plugin loaded: %s", p.name)
            return p
        return None

    def load(self, plugin_type):
        try:
            p = self._load_if_enabled(plugin_type)

            if p is not None:
                # Order of execution is significant!
                p.create_actions()
                p.create_menu()
                p.create_tools()
                p.create_toolbars()
                p.create_widgets()
        except Exception:
            self.warn(sys._getframe().f_code.co_name, plugin_type)

    def load_from_file(self, path):
        master = Plugin
        prev = self._inheritors(master)
        name = os.path.splitext(os.path.basename(path))[0]
        mod_name = 'hyperspyui.plugins.' + name
        reload_plugins = mod_name in sys.modules
        try:
            if sys.version_info >= (3, 5):
                spec = importlib.util.spec_from_file_location(
                    mod_name, path)
                mod = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(mod)
            else:
                importlib.machinery.SourceFileLoader(
                    mod_name, path).load_module()
        except Exception:
            self.warn("import", name)
        loaded = self._inheritors(master).difference(prev)

        new_ps = []
        for plug_type in loaded:
            if reload_plugins and plug_type.name in self.plugins:
                # Unload any plugins with same name
                self.unload(self.plugins[plug_type.name])
            try:
                p = self._load_if_enabled(plug_type)
            except Exception:
                self.warn('load', plug_type.name)
            if p is not None:
                new_ps.append(p)
        for p in new_ps:
            try:
                p.create_actions()
            except Exception:
                self.warn('create_actions', p.name)
        for p in new_ps:
            try:
                p.create_menu()
            except Exception:
                self.warn('create_menu', p.name)
        for p in new_ps:
            try:
                p.create_tools()
            except Exception:
                self.warn('create_tools', p.name)
        for p in new_ps:
            try:
                p.create_toolbars()
            except Exception:
                self.warn('create_toolbars', p.name)
        for p in new_ps:
            try:
                p.create_widgets()
            except Exception:
                self.warn('create_widgets', p.name)
        return new_ps

    def unload(self, plugin):
        plugin.unload()
        self.plugins.pop(plugin.name)

    def reload(self, plugin):
        new_module = importlib.reload(sys.modules[plugin.__module__])
        new_ptype = new_module[plugin.__class__.__name__]
        if new_ptype is not None:
            self.unload(plugin)
            self.load(new_ptype)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Tue Nov 25 02:10:29 2014

@author: Vidar Tonaas Fauske
"""

from distutils.version import LooseVersion
import os
import platform
import sys
import locale
from contextlib import contextmanager


# TODO: Make sure tools are disconnected when closing signal!

# TODO: Autohide toolbars w.r.t. signal type. (maybe?)
# TODO: Add selection groups for signals by hierarchy in DataViewWidget
# TODO: Make sure everything records / plays back code
# TODO: MPL artist editor via traitsui with custom handler
# TODO: Batch processing dialog (browse + drop&drop target)
# TODO: Editor utils: threading + parallell processing (w/batch input)
# TODO: Add xray 3rd state: Element present, but not used for maps
# TODO: Add xray background window tool for modifying those by auto
# TODO: Add quantification button. Needs k-factor setup.
# TODO: Create contributed plugins repository with UI integrated hot-load
# TODO: Auto-resize font in Periodic table
# TODO: Utilities to combine signals into stack and reverse
# TODO: Make editor tabs drag&droppable + default to all in one
# TODO: Make data type changer handle RGBX data types.
# TODO: Variable explorer (Signal/Model etc.). 3rd party library?
# TODO: Disable tools when no figure


import logging
logging.basicConfig()
# Note that this might be overriden during hyperspy import
LOGLEVEL = logging.DEBUG
logging.getLogger('hyperspy').setLevel(LOGLEVEL)
logging.getLogger('hyperspyui').setLevel(LOGLEVEL)


def _get_logfile():
    _, exe_name = os.path.split(sys.executable)
    if exe_name.startswith('pythonw'):
        log_path = os.path.dirname(__file__) + '/hyperspyui.log'
        try:
            log_file = open(log_path, 'w')
        except IOError:
            log_path = os.path.expanduser('~/hyperspyui.log')
            try:
                log_file = open(log_path, 'w')
            except IOError:
                log_file = None     # No log file for us!
    else:
        log_file = None
    return log_file


def get_splash():
    from qtpy.QtCore import Qt
    from qtpy.QtWidgets import QApplication, QSplashScreen
    from qtpy.QtGui import QColor, QPixmap
    splash_pix = QPixmap(os.path.join(
        os.path.dirname(__file__), 'images', 'splash.png'))
    splash = QSplashScreen(splash_pix, Qt.WindowStaysOnTopHint)
    splash.show()
    splash.showMessage("Initializing...", Qt.AlignBottom | Qt.AlignCenter |
                       Qt.AlignAbsolute, QColor(Qt.white))
    QApplication.processEvents()
    return splash


def main():
    from qtpy.QtCore import Qt, QCoreApplication, qVersion
    from qtpy.QtWidgets import QApplication
    from qtpy import API

    # Fixes issues with Big Sur
    # https://bugreports.qt.io/browse/QTBUG-87014, fixed in qt 5.15.2
    if (sys.platform == 'darwin' and
            LooseVersion(platform.mac_ver()[0]) >= LooseVersion("10.16") and
            LooseVersion(qVersion()) < LooseVersion("5.15.2") and
            "QT_MAC_WANTS_LAYER" not in os.environ):
        os.environ["QT_MAC_WANTS_LAYER"] = "1"

    from hyperspyui.version import __version__
    from hyperspyui.settings import Settings

    # QtWebEngineWidgets must be imported before a QCoreApplication instance
    # is created (used in eelsdb plugin)
    # Avoid a bug in Qt: https://bugreports.qt.io/browse/QTBUG-46720
    from qtpy import QtWebEngineWidgets

    # Need to set early to make QSettings accessible
    QCoreApplication.setApplicationName("HyperSpyUI")
    QCoreApplication.setOrganizationName("Hyperspy")
    QCoreApplication.setApplicationVersion(__version__)
    # To avoid the warning:
    # Qt WebEngine seems to be initialized from a plugin. Please set
    # Qt::AA_ShareOpenGLContexts using QCoreApplication::setAttribute before
    # constructing QGuiApplication.
    # Only available for pyqt>=5.4
    if hasattr(Qt, "AA_ShareOpenGLContexts"):
        QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts)

    # First, clear all default settings!
    # TODO: This will cause a concurrency issue with multiple launch
    Settings.clear_defaults()
    # Setup default for single/multi-instance
    settings = Settings(group="General")
    settings.set_default('allow_multiple_instances', False)
    if settings['allow_multiple_instances', bool]:
        # Using multiple instances, get a new application
        app = QApplication(sys.argv)
    else:
        # Make sure we only have a single instance
        from hyperspyui.singleapplication import get_app
        app = get_app('hyperspyui')

    splash = get_splash()

    log_file = _get_logfile()
    if log_file:
        sys.stdout = sys.stderr = log_file
    else:
        @contextmanager
        def dummy_context_manager(*args, **kwargs):
            yield
        log_file = dummy_context_manager()

    with log_file:
        # Need to have import here, since QApplication needs to be called first
        from hyperspyui.mainwindow import MainWindow

        form = MainWindow(splash=splash)
        if not settings['allow_multiple_instances', bool]:
            if "pyqt" in API:
                app.messageAvailable.connect(form.handleSecondInstance)
            elif API == 'pyside':
                app.messageReceived.connect(form.handleSecondInstance)
        form.showMaximized()

        form.splash.hide()
        form.load_complete.emit()
        # Ensure logging is OK
        import hyperspy.api as hs
        hs.set_log_level(LOGLEVEL)

        app.exec_()


if __name__ == "__main__":
    locale.setlocale(locale.LC_ALL, '')
    sys.path.append(os.path.dirname(__file__))
    main()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Thu Jul 30 11:42:05 2015

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore
import numpy as np

from hyperspy.drawing.widgets import Line2DWidget, RangeWidget
from hyperspy.roi import BaseInteractiveROI, SpanROI, Line2DROI

from hyperspyui.log import logger
from hyperspyui.tools import SignalFigureTool
from hyperspyui.util import crosshair_cursor


class LineTool(SignalFigureTool):

    """
    Tool to select a ROI on a  signal interactively. Simply click and drag in a
    figure to create an ROI, and then press enter to indicate that the
    selection is complete, or ESC to cancel. The selection can also be aborted
    by simply selecting a different tool.
    """

    accepted = QtCore.Signal([BaseInteractiveROI],
                             [BaseInteractiveROI, SignalFigureTool])
    updated = QtCore.Signal([BaseInteractiveROI],
                            [BaseInteractiveROI, SignalFigureTool])
    cancelled = QtCore.Signal()

    def __init__(self, windows=None):
        super(LineTool, self).__init__(windows)
        self.widget2d = Line2DWidget(None)
        self.widget2d.set_on(False)
        self.widget1d = RangeWidget(None)
        self.widget1d.set_on(False)
        self.valid_dimensions = [1, 2]
        self.axes = None

    @property
    def widget(self):
        if self.ndim == 1:
            return self.widget1d
        else:
            return self.widget2d

    @property
    def ndim(self):
        if self.axes is None:
            return 0
        return len(self.axes)

    def is_on(self):
        return self.widget.is_on()

    def in_ax(self, ax):
        return ax == self.widget.ax

    def get_name(self):
        return "Line tool"

    def get_category(self):
        return 'Signal'

    def make_cursor(self):
        return crosshair_cursor()

    def is_selectable(self):
        return True

    def on_mousedown(self, event):
        # Only accept mouse down inside axes
        if event.inaxes is None:
            return

        axes = self._get_axes(event)
        if not axes:
            logger.warning('Line tool only works on HyperSpy signal plots!')
            return
        # Make sure we have a figure with valid dimensions
        if len(axes) not in self.valid_dimensions:
            return
        self.axes = axes
        # If we already have a widget, make sure dragging is passed through
        if self.is_on():
            if any([p.contains(event)[0] is True for p in self.widget.patch]):
                return              # Interacting, handle in widget
            # Cancel previous and start new
            self.cancel()
            # Cancel reset axes, so set again
            self.axes = axes

        # Find out which axes of Signal are plotted in figure
        s = self._get_signal(event.inaxes.figure)
        if s is None:
            logger.warning('Line tool only works on HyperSpy signal plots!')
            return
        am = s.axes_manager

        self.widget.axes = axes
        if self.ndim > 1:
            self.widget.axes_manager = am

        # Have what we need, create widget
        x, y = event.xdata, event.ydata
        self.widget.set_mpl_ax(event.inaxes)  # connects
        new_widget = not self.is_on()
        if self.ndim == 1:
            self.widget.position = (x,)
            self.widget.size = np.array([0])
            self.widget.set_on(True)
            span = self.widget.span
            span.buttonDown = True
            span.on_move_cid = \
                span.canvas.mpl_connect('motion_notify_event',
                                        span.move_right)
        else:
            self.widget.position = np.array(((x, y), (x, y)))
            self.widget.size = np.array([0])
            self.widget.picked = True
            self.widget.func = Line2DWidget.FUNC_B | Line2DWidget.FUNC_RESIZE
            self.widget._drag_start = [x, y]
            self.widget._drag_store = (self.widget.position, self.widget.size)
            self.widget.set_on(True)
        if new_widget:
            self.widget.events.changed.connect(self._on_change,
                                               {'obj': 'widget'})

    def on_keyup(self, event):
        if event.key == 'enter':
            self.accept()
        elif event.key == 'escape':
            self.cancel()

    def _on_change(self, widget):
        if self.is_on():
            if self.ndim == 1:
                roi = SpanROI(0, 1)
            elif self.ndim > 1:
                roi = Line2DROI(0, 0, 1, 1, 1)
            roi._on_widget_change(self.widget)  # ROI gets coords from widget
            self.updated[BaseInteractiveROI].emit(roi)
            self.updated[BaseInteractiveROI, SignalFigureTool].emit(roi, self)

    def accept(self):
        if self.is_on():
            if self.ndim == 1:
                roi = SpanROI(0, 1)
            elif self.ndim > 1:
                roi = Line2DROI(0, 0, 1, 1, 1)
            roi._on_widget_change(self.widget)  # ROI gets coords from widget
            self.accepted[BaseInteractiveROI].emit(roi)
            self.accepted[BaseInteractiveROI, SignalFigureTool].emit(roi, self)

    def cancel(self):
        if self.widget.is_on():
            self.widget.set_on(False)
            self.widget.size = np.array([0])    # Prevents flickering
        if self._on_change in self.widget.events.changed.connected:
            self.widget.events.changed.disconnect(self._on_change)
        self.axes = None
        self.cancelled.emit()

    def disconnect_windows(self, windows):
        super(LineTool, self).disconnect_windows(windows)
        self.cancel()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from hyperspyui.plugins.plugin import Plugin
import numpy as np
import hyperspy.api as hs
from hyperspy.drawing import utils
from functools import partial


class VirtualBfDf(Plugin):
    name = "Virtual BF/DF"

    def __init__(self, *args, **kwargs):
        super(VirtualBfDf, self).__init__(*args, **kwargs)
        self._rois = []

    def create_actions(self):
        self.add_action(self.name + '.virtual_navigator',
                        "Virtual navigator",
                        self.virtual_navigator,
                        tip="Set the navigator inesity by a virtual aperture")
        self.add_action(self.name + '.virtual_aperture',
                        "Virtual aperture",
                        self.virtual_aperture,
                        tip="Add a virtual aperture to the diffraction image")
        self.add_action(self.name + '.virtual_annulus',
                        "Virtual annulus",
                        self.virtual_annulus,
                        tip="Add a virtual annulus to the diffraction image")

    def create_menu(self):
        self.add_menuitem(
            'Diffraction', self.ui.actions[self.name + '.virtual_navigator'])
        self.add_menuitem(
            'Diffraction', self.ui.actions[self.name + '.virtual_aperture'])
        self.add_menuitem(
            'Diffraction', self.ui.actions[self.name + '.virtual_annulus'])

    def _on_close(self, roi):
        for w in roi.widgets:
            w.close()
        roi.signal_map.clear()
        if roi in self._rois:
            self._rois.remove(roi)

    def virtual_navigator(self, signal=None):
        return self.virtual_aperture(signal=signal, navigate=True)

    def virtual_annulus(self, signal=None):
        return self.virtual_aperture(signal=signal, annulus=True)

    def virtual_aperture(self, signal=None, annulus=False, navigate=False):
        ui = self.ui
        if signal is None:
            signal = ui.get_selected_signal()
        dd = np.array([a.high_value + a.low_value for a in
                       signal.axes_manager.signal_axes]) / 2.0
        r = hs.roi.CircleROI(dd[0], dd[1],
                             signal.axes_manager.signal_axes[0].scale*3)
        if annulus:
            r.r_inner = signal.axes_manager.signal_axes[0].scale*2
        s_virtual = r.interactive(signal, None,
                                  axes=signal.axes_manager.signal_axes)
        s_nav = hs.interactive(
            s_virtual.mean,
            s_virtual.events.data_changed,
            axis=s_virtual.axes_manager.signal_axes)
        s_nav.axes_manager.set_signal_dimension(2)
        if navigate:
            signal.plot(navigator=s_nav)
            signal._plot.navigator_plot.update()
            s_nav.events.data_changed.connect(
                signal._plot.navigator_plot.update, [])
            utils.on_figure_window_close(
                signal._plot.navigator_plot.figure,
                partial(self._on_close, r))
        else:
            s_nav.plot()
            utils.on_figure_window_close(
                s_nav._plot.signal_plot.figure,
                partial(self._on_close, r))

        if navigate:
            r.add_widget(signal, axes=signal.axes_manager.signal_axes,
                         color='darkorange')
        else:
            r.add_widget(signal, axes=signal.axes_manager.signal_axes)
        self._rois.append(r)
        self.record_code("<p>.virtual_aperture(navigate=%s)" % navigate)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sat Feb 21 16:05:41 2015

@author: Vidar Tonaas Fauske
"""

from .mainwindowbase import MainWindowBase, tr

import os
import inspect
from functools import partial

from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialogButtonBox

from hyperspyui.smartcolorsvgiconengine import SmartColorSVGIconEngine
from hyperspyui.advancedaction import AdvancedAction

from hyperspyui.log import logger as _logger

from hyperspyui import hooktraitsui

hooktraitsui.hook_traitsui()


class MainWindowUtils(MainWindowBase):
    """
    Adds UI utility functions to the main window, including traitsui dialog
    capture.
    """

    def __init__(self, parent=None):
        # traitsui backend bindings
        hooktraitsui.connect_created(self.on_traits_dialog)
        hooktraitsui.connect_destroyed(self.on_traits_destroyed)

        super(MainWindowUtils, self).__init__(parent)


    # --------- traitsui Events ---------

    def capture_traits_dialog(self, callback):
        self.should_capture_traits = callback

    def on_traits_dialog(self, dialog, ui, parent):
        self.traits_dialogs.append(dialog)
        if parent is None:
            if self.should_capture_traits:
                self.should_capture_traits(dialog)
                self.should_capture_traits = None
            else:
                dialog.setParent(self, QtCore.Qt.Tool)
                dialog.show()
                dialog.activateWindow()

    def on_traits_destroyed(self, dialog):
        if dialog in self.traits_dialogs:
            self.traits_dialogs.remove(dialog)

    # --------- End traitsui Events ---------

    def set_status(self, msg):
        """
        Display 'msg' in window's statusbar.
        """
        # TODO: What info is needed? Add simple label first, create utility to
        # add more?
        self.statusBar().showMessage(msg)

    def _make_action(self, label, icon, shortcut, tip):
        ac = AdvancedAction(tr(label), self)
        if icon:
            ac.setIcon(self.make_icon(icon))
        if shortcut is not None:
            ac.setShortcut(shortcut)
        if tip is not None:
            ac.setStatusTip(tr(tip))
        return ac

    def _wire_action(self, ac, key, callback, selection_callback):
        try:
            keywords = inspect.getfullargspec(callback).args
        except TypeError:
            keywords = None
        if keywords and 'advanced' in keywords:
            orig_callback = callback

            def wrap(checked, advanced):
                orig_callback(advanced=advanced)
            callback = wrap
            ac.triggered[bool, bool].connect(callback)
        else:
            ac.triggered.connect(callback)
        # Use docstring for action
        if callback.__doc__:
            d = callback.__doc__
            if d.startswith('partial('):
                # Fix docstring of partial functions:
                d = callback.func.__doc__
            ac.__doc__ = d
        if selection_callback is not None:
            self._action_selection_cbs[key] = selection_callback
            ac.setEnabled(False)


    def add_action(self, key, label, callback, tip=None, icon=None,
                   shortcut=None, selection_callback=None):
        """
        Create and add a QAction to self.actions[key]. 'label' is used as the
        short description of the action, and 'tip' as the long description.
        The tip is typically shown in the statusbar. The callback is called
        when the action is triggered(). The optional 'icon' should either be a
        QIcon, or a path to an icon file, and is used to depict the action on
        toolbar buttons and in menus.

        If `selection_callback` is supplied, it is called whenever the
        currently selected signal/figure changes. This allows the callback to
        enable/disable the action to reflect whether the selected figure/signal
        is supported for the action.
        """
        ac = self._make_action(label, icon, shortcut, tip)
        self._wire_action(ac, key, callback, selection_callback)
        self.actions[key] = ac
        return ac

    def add_toolbar_button(self, category, action):
        """
        Add the supplied 'action' as a toolbar button. If the toolbar defined
        by 'cateogry' does not exist, it will be created in
        self.toolbars[category].
        """
        if category in self.toolbars:
            tb = self.toolbars[category]
        else:
            tb = QtWidgets.QToolBar(tr(category) + tr(" toolbar"), self)
            tb.setObjectName(category + "_toolbar")
            self.addToolBar(Qt.LeftToolBarArea, tb)
            self.toolbars[category] = tb

        if not isinstance(action, QtWidgets.QAction):
            action = self.actions[action]
        tb.addAction(action)

    def remove_toolbar_button(self, category, action):
        tb = self.toolbars[category]
        tb.removeAction(action)
        if len(tb.actions()) < 1:
            self.removeToolBar(tb)

    def add_menuitem(self, category, action, label=None):
        """
        Add the supplied 'action' as a menu entry. If the menu defined
        by 'cateogry' does not exist, it will be created in
        self.menus[category].

        If the label argument is not supplied, category will be used.
        """
        if category in self.menus:
            m = self.menus[category]
        else:
            if label is None:
                label = category
            # Make sure we add menu before window menu
            if self.windowmenu is None:
                m = self.menuBar().addMenu(label)
            else:
                m = QtWidgets.QMenu(label)
                self.menuBar().insertMenu(self.windowmenu.menuAction(), m)
            self.menus[category] = m

        if not isinstance(action, QtWidgets.QAction):
            action = self.actions[action]
        m.addAction(action)

    def add_tool(self, tool, selection_callback=None):
        if isinstance(tool, type):
            t = tool(self.figures)
            key = tool.__name__
        else:
            t = tool
            try:
                key = t.get_name()
            except NotImplementedError:
                key = tool.__class__.__name__
        self.tools.append(t)
        if t.single_action() is not None:
            self.add_action(key, t.get_name(), t.single_action(),
                            selection_callback=selection_callback,
                            icon=t.get_icon(), tip=t.get_description())
            self.add_toolbar_button(t.get_category(), self.actions[key])
        elif t.is_selectable():
            f = partial(self.select_tool, t)
            self.add_action(key, t.get_name(), f, icon=t.get_icon(),
                            selection_callback=selection_callback,
                            tip=t.get_description())
            self.selectable_tools.addAction(self.actions[key])
            self.actions[key].setCheckable(True)
            self.add_toolbar_button(t.get_category(), self.actions[key])
        return key

    def remove_tool(self, tool):
        if isinstance(tool, type):
            for t in self.tools:
                if isinstance(t, tool):
                    break
            key = tool.__name__
        else:
            t = tool
            try:
                key = t.get_name()
            except NotImplementedError:
                key = tool.__class__.__name__
        self.tools.remove(t)
        ac = self.actions.pop(key, None)
        if ac is not None:
            self.remove_toolbar_button(t.get_category(), ac)
            if t.is_selectable():
                self.selectable_tools.removeAction(ac)

    def add_widget(self, widget, floating=None):
        """
        Add the passed 'widget' to the main window. If the widget is not a
        QDockWidget, it will be wrapped into one. The QDockWidget is returned.
        The widget is also added to the window menu self.windowmenu, so that
        it's visibility can be toggled.

        The parameter 'floating' specifies whether the widget should be made
        floating. If None, the value of the setting 'default_widget_floating'
        is used.
        """
        if floating is None:
            floating = self.settings['default_widget_floating', bool]
        if isinstance(widget, QtWidgets.QDockWidget):
            d = widget
        else:
            d = QtWidgets.QDockWidget(self)
            d.setWidget(widget)
            d.setWindowTitle(widget.windowTitle())
        if not d.objectName():
            d.setObjectName(d.windowTitle())
        d.setAllowedAreas(Qt.RightDockWidgetArea | Qt.LeftDockWidgetArea)
        self.addDockWidget(Qt.RightDockWidgetArea, d)
        d.setFloating(floating)

        self.widgets.append(widget)

        # Insert widgets in Windows menu before separator (figures are after)
        self.windowmenu.insertAction(self.windowmenu_sep, d.toggleViewAction())
        return d

    def make_icon(self, icon):
        """
        Create an icon that coheres to the internal standard for icons.

        Parameters
        ----------
        icon : {string | QIcon}
            If icon is a path, it loads the file. If the path does not
            correspond to a valid file, it is checked if it is a valid
            path relative to the 'images' folder of the package.
            After loading, SVG files will be run through
            `SmartColorSVGIconEngine` to adapt suitable icons to the
            current palette. If a QIcon is passed directly, it is also
            sent through `SmartColorSVGIconEngine`.
        """
        if not isinstance(icon, QtGui.QIcon):
            if isinstance(icon, str) and not os.path.isfile(icon):
                sugg = os.path.dirname(__file__) + '/images/' + icon
                if os.path.isfile(sugg):
                    icon = sugg
            if isinstance(icon, str) and (
                    icon.endswith('svg') or
                    icon.endswith('svgz') or
                    icon.endswith('svg.gz')):
                ie = SmartColorSVGIconEngine()
                path = icon
                icon = QtGui.QIcon(ie)
                icon.addFile(path)
            else:
                icon = QtGui.QIcon(icon)
        else:
            icon = QtGui.QIcon(SmartColorSVGIconEngine(icon))
        return icon

    def prompt_files(self, extension_filter=None, path=None, exists=True,
                     title=None, def_filter=None):
        if title is None:
            title = tr('Load file') if exists else tr('Save file')
        path = path or self.cur_dir or ''
        if def_filter is None and extension_filter:
            def_filter = extension_filter.split(';;', maxsplit=1)[0]

        if exists:
            filenames = QtWidgets.QFileDialog.getOpenFileNames(
                self, title, path, extension_filter)
        else:

            filenames = QtWidgets.QFileDialog.getSaveFileName(
                self, title, path, extension_filter, def_filter)
        # Pyside returns tuple, PyQt not
        if isinstance(filenames, tuple):
            filenames = filenames[0]
        return filenames

    def get_figure_filepath_suggestion(self, figure, deault_ext=None):
        """
        Get a suggestion for a file path for saving `figure`.
        """
        canvas = figure.widget()
        if deault_ext is None:
            deault_ext = canvas.get_default_filetype()

        f = canvas.get_default_filename()
        if not f:
            f = self.cur_dir

        # Analyze suggested filename
        base, tail = os.path.split(f)
        fn, ext = os.path.splitext(tail)
        _logger.debug('fn before cleaning is: {}'.format(fn))

        # Remove illegal characters and newlines from filename:
        reserved_characters = r'<>:"/\|?*'
        for c in reserved_characters:
            fn = fn.replace(c, '')
        fn = fn.replace('\n', ' ')

        _logger.debug('fn after cleaning is: {}'.format(fn))

        # If no directory in filename, use self.cur_dir's dirname
        if base is None or base == "":
            base = os.path.dirname(self.cur_dir)
        # If extension is not valid, use the defualt
        if ext not in canvas.get_supported_filetypes():
            ext = deault_ext

        # Build suggestion and return
        path_suggestion = os.path.sep.join((base, fn))
        path_suggestion = os.path.extsep.join((path_suggestion, ext))
        return path_suggestion

    def save_figure(self, figure=None):
        """
        Save the matplotlib figure. If a figure is not passed, it tries to
        save whichever is active (using `activeSubWindow()` of the MDI area).
        """
        if figure is None:
            figure = self.main_frame.activeSubWindow()
            if figure is None:
                return
        path_suggestion = self.get_figure_filepath_suggestion(figure)
        canvas = figure.widget()

        # Build type selection string
        def_type = os.path.extsep + canvas.get_default_filetype()
        extensions = canvas.get_supported_filetypes_grouped()
        type_choices = "All types (*.*)"
        for group, exts in extensions.items():
            fmt = group + \
                ' (' + \
                '; '.join(["*" + os.path.extsep + sube for sube in exts]) + ')'
            type_choices = ';;'.join((type_choices, fmt))
            if def_type[1:] in exts:
                def_type = fmt

        # Present filename prompt
        filename = self.prompt_files(type_choices, path_suggestion,
                                     exists=False, def_filter=def_type)
        if filename:
            canvas.figure.savefig(filename)

    def show_okcancel_dialog(self, title, widget, modal=True):
        """
        Show a dialog with the passed widget and OK and cancel buttons.
        """
        diag = QtWidgets.QDialog(self)
        diag.setWindowTitle(title)
        diag.setWindowFlags(Qt.Tool)

        btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                                Qt.Horizontal, diag)
        btns.accepted.connect(diag.accept)
        btns.rejected.connect(diag.reject)

        box = QtWidgets.QVBoxLayout(diag)
        box.addWidget(widget)
        box.addWidget(btns)
        diag.setLayout(box)

        if modal:
            diag.exec_()
        else:
            diag.show()
        # Return the dialog for result checking, and to keep widget in scope
        # for caller
        return diag


class MainWindowActionRecorder(MainWindowUtils):

    """
    Adds recorder functionality.
    """

    def __init__(self, parent=None):
        self.recorders = []
        super(MainWindowActionRecorder, self).__init__(parent)

    def _wire_action(self, ac, key, callback, selection_callback):
        # Connect monitor
        ac.triggered.connect(partial(self.record_action, key))
        # Wire as normal
        super(MainWindowActionRecorder, self)._wire_action(
            ac, key, callback, selection_callback)

    def record_action(self, key):
        for r in self.recorders:
            r.add_action(key)

    def record_code(self, code):
        for r in self.recorders:
            r.add_code(code)

    def on_console_executing(self, source):
        super(MainWindowActionRecorder, self).on_console_executing(source)
        self.record_code(source)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Oct 27 23:17:25 2014

@author: Vidar Tonaas Fauske
"""


from qtpy import QtWidgets


class BindingList(list):

    """
    A list that has been extended to sync other lists or collections to changes
    in its contents. By custom targets, it can also be used to trigger events
    on addition/removal. Only append and remove actions are required, as
    extend, insert and pop can be inferred (insert loses order however), but
    for reasons of speed, it is recommended to supply all if they are
    available. Supported targets for add_target are types 'list' and
    'QListWidget'.
    """

    def __init__(self, target=None, *args, **kwargs):
        super(BindingList, self).__init__(*args, **kwargs)
        self.set_target(target)

    def set_target(self, target):
        self.targets = {}
        self.add_target(target)

    def add_custom(self, target, append, insert, extend, remove, pop):
        cb = {'ap': append, 'in': insert,
              'ex': extend, 're': remove, 'po': pop}
        self.targets[target] = cb

    def add_target(self, target):
        if target is None:
            return
        elif isinstance(target, list):
            cb = {'ap': target.append, 'in': target.insert,
                  'ex': target.extend, 're': target.remove, 'po': target.pop}
#        elif isinstance(target, QList):
#            cb = {'ap': target.append, 'in': target.insert,
#                  'ex': target.append, 're': target.removeOne,
#                  'po': target.removeAt}
        elif isinstance(target, QtWidgets.QListWidget):
            def qlr(value):
                target.takeItem(self.index(value))
            cb = {'ap': target.addItem, 'in': target.insertItem,
                  'ex': target.addItems, 're': qlr,
                  'po': target.takeItem}
        self.targets[target] = cb

    def remove_target(self, target):
        self.targets.pop(target, 0)

    def append(self, object):
        super(BindingList, self).append(object)
        for t in list(self.targets.values()):
            if t['ap'] is not None:
                t['ap'](object)

    def insert(self, index, object):
        super(BindingList, self).insert(index, object)
        for t in list(self.targets.values()):
            if t['in'] is not None:
                t['in'](index, object)
            elif t['ap'] is not None:
                t['ap'](object)

    def extend(self, iterable):
        super(BindingList, self).extend(iterable)
        for t in list(self.targets.values()):
            if t['ex'] is not None:
                t['ex'](iterable)
            if t['ap'] is not None:
                for v in iterable:
                    t['ap'](v)

    def remove(self, value):
        if value not in self:
            return

        for t in list(self.targets.values()):
            if t['re'] is not None:
                t['re'](value)
        super(BindingList, self).remove(value)

    def pop(self, index=-1):
        if index < 0:
            index = len(self) + index
        for t in list(self.targets.values()):
            if t['po'] is not None:
                t['po'](index)
            elif t['re'] is not None:
                v = self[index]
                t['re'](v)
        return super(BindingList, self).pop(index)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Nov 21 22:11:04 2014

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore, QtWidgets

import numpy as np


def tr(text):
    return QtCore.QCoreApplication.translate("ExtendedQWidgets", text)


class ExToolWindow(QtWidgets.QDialog):

    """
    QDialog with Qt.Tool window flags.
    """

    def __init__(self, parent=None):
        super(ExToolWindow, self).__init__(parent)
        self.setWindowFlags(QtCore.Qt.Tool)


class FigureWidget(QtWidgets.QDockWidget):

    def __init__(self, main_window, parent=None):
        super(FigureWidget, self).__init__(parent)
        self.ui = main_window
        self._last_window = None

        self.connect_()
        self.visibilityChanged.connect(self.on_visibility)

    def connect_(self, action=None):
        """
        Connects the widget to its update trigger, which is either a supplied
        action, or by default, the subWindowActivated of the MainWindow.
        """
        if action is None:
            action = self.ui.main_frame.subWindowActivated
        action.connect(self._on_figure_change)

    def disconnect(self, action=None):
        """
        Disconnects an update trigger connected with connect().
        """
        if action is None:
            action = self.ui.main_frame.subWindowActivated
        action.disconnect(self.on_change)

    def _on_figure_change(self, window):
        """
        Called when a connected update triggers. If the window is valid, it
        sets up the traitsui dialog capture, and calls cb_make_dialog.  If the
        window is invalid, it clears the widget.
        """
        self._last_window = window

    def on_visibility(self, visible):
        if visible:
            self._on_figure_change(self._last_window)


class ExClickLabel(QtWidgets.QLabel):

    """
    QLabel with 'clicked()' signal.
    """
    clicked = QtCore.Signal()

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.clicked.emit()
        super(ExClickLabel, self).mouseReleaseEvent(event)


class ExMessageBox(QtWidgets.QMessageBox):

    def isChecked(self):
        if self.cb is None:
            raise AttributeError
        return self.cb.isChecked()

    def setCheckBox(self, cb):
        try:
            super(ExMessageBox, self).setCheckBox(cb)
        except AttributeError:
            oldcb = self.checkBox()
            if oldcb is not None:
                self.removeButton(oldcb)
            self._checkBox = cb
            if cb is not None:
                cb.blockSignals(True)
                self.addButton(cb, QtWidgets.QMessageBox.ResetRole)

    def checkBox(self):
        try:
            return super(ExMessageBox, self).checkBox()
        except AttributeError:
            pass
        try:
            return self._checkBox
        except AttributeError:
            return None


class ExRememberPrompt(ExMessageBox):

    def __init__(self, *args, **kwargs):
        super(ExRememberPrompt, self).__init__(*args, **kwargs)
        self.cb = QtWidgets.QCheckBox(tr("Remember this choice"), self)
        self.setCheckBox(self.cb)


class ExDoubleSlider(QtWidgets.QSlider):

    """
    QSlider with double values instead of int values.
    """
    double_valueChanged = QtCore.Signal(float)

    def __init__(self, parent=None, orientation=None):
        if orientation is None:
            super(ExDoubleSlider, self).__init__(parent)
        else:
            super(ExDoubleSlider, self).__init__(orientation, parent)
        self.steps = 1000
        self._range = (0.0, 1.0)
        self.valueChanged.connect(self._on_change)

    def setRange(self, vmin, vmax):
        if isinstance(vmin, (np.complex64, np.complex128)):
            vmin = np.abs(vmin)
        if isinstance(vmax, (np.complex64, np.complex128)):
            vmax = np.abs(vmax)
        self._range = (vmin, vmax)
        return super(ExDoubleSlider, self).setRange(0, self.steps)

    def setValue(self, value):
        vmin, vmax = self._range
        if isinstance(value, (np.complex64, np.complex128)):
            value = np.abs(value)
        try:
            v = int((value - vmin) * self.steps / (vmax - vmin))
        except (ZeroDivisionError, OverflowError, ValueError):
            v = 0
            self.setEnabled(False)
        return super(ExDoubleSlider, self).setValue(v)

    def value(self):
        v = super(ExDoubleSlider, self).value()
        return self._int2dbl(v)

    def _int2dbl(self, intval):
        vmin, vmax = self._range
        return vmin + intval * (vmax - vmin) / self.steps

    def _on_change(self, intval):
        dblval = self._int2dbl(intval)
        self.double_valueChanged.emit(dblval)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Wed Jul 29 18:09:56 2015

@author: vidarton
"""


from qtpy import QtGui, QtCore, QtSvg, QtWidgets
from qtpy.QtGui import QPalette

import re


class SmartColorSVGIconEngine(QtGui.QIconEngine):
    """
    This class is basically a port to Python from the code for Qt's
    QSvgIconEnginePlugin. On top of this has been added the ability to exchange
    colors in the icons. It's also possible in the future to do more advanced
    XML pro-processing. Note: Some of the more esoteric uses of QIcon might
    not be fully tested for yet.
    """

    serial_gen = 0

    def __init__(self, use_qt_disabled=False, other=None):
        if other:
            super(SmartColorSVGIconEngine, self).__init__(other)
            if isinstance(other, SmartColorSVGIconEngine):
                self._svgFiles = other._svgFiles
                if other._svgBuffers:
                    self._svgBuffers = other._svgBuffers.copy()
                if other._addedPixmaps:
                    self._addedPixmaps = other._addedPixmaps.copy()
        else:
            super(SmartColorSVGIconEngine, self).__init__()
        self._addedPixmaps = {}
        self._svgFiles = {}
        self._svgBuffers = {}
        self.default_key = (QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.custom_color_replacements = {}
        self._automatic_color_replacements = {}
        self.use_qt_disabled = use_qt_disabled
        self._palette_key = None
        self._set_replacements_from_palette()

    def _set_replacements_from_palette(self):
        """Sets `_automatic_color_replacements` attribute from application
        palette. Updates `_palette_key` to the cacheKey of the used palette.
        """
        palette = QtWidgets.QApplication.palette()
        foreground = palette.color(QPalette.Active, QPalette.WindowText)
        background = palette.color(QPalette.Active, QPalette.Window)
        disabled_foreground = palette.color(QPalette.Disabled,
                                            QPalette.WindowText)
        disabled_background = palette.color(QPalette.Disabled,
                                            QPalette.Window)
        self._automatic_color_replacements = {
            'default': {
                b'#000000': foreground.name().encode('ascii'),
                b'black': foreground.name().encode('ascii'),
                b'#ffffff': background.name().encode('ascii'),
                b'white': background.name().encode('ascii'),
                },
            'disabled': {
                b'#000000': disabled_foreground.name().encode('ascii'),
                b'black': disabled_foreground.name().encode('ascii'),
                b'#ffffff': disabled_background.name().encode('ascii'),
                b'white': disabled_background.name().encode('ascii'),
                }
            }
        self._palette_key = palette.cacheKey()

    def _make_cache_key(self, size, mode, state):
        """Make a unique, reproducable key to store/lookup icons in the pixmap
        cache.
        """
        return str(self) + str((size, mode, state))

    def _replace_in_stream(self, filename, color_key='default'):
        """Opens supplied SVG file, parses the file replacing colors according
        to the dictionary `color_replacements`. Returns a QByteArray of the
        processed content.
        """
        f_low = filename.lower()
        if f_low.endswith('.svgz') or f_low.endswith('.svg.gz'):
            import gzip
            with gzip.open(filename, 'rb') as svg_file:
                svg_cnt = svg_file.read()
        else:
            with open(filename, 'rb') as svg_file:
                svg_cnt = svg_file.read()

        # Merge automatic and custom LUTs
        color_table = self._automatic_color_replacements.copy()
        color_table.update(self.custom_color_replacements)
        if color_key not in color_table:
            color_key = 'default'

        # Perform replacement according to correct table
        for old, new in color_table[color_key].items():
            re_exp = re.compile(re.escape(old), re.IGNORECASE)
            svg_cnt = re_exp.sub(new, svg_cnt)
        out = QtCore.QByteArray(svg_cnt)
        return out

    def _loadDataForModeAndState(self, renderer, mode, state):
        """Load SVG data to renderer.
        """
        # First, try to load from a buffer if available.
        if (mode, state) in self._svgBuffers:
            buf = self._svgBuffers[(mode, state)]
        elif self.default_key in self._svgBuffers:
            buf = self._svgBuffers[self.default_key]
        else:
            buf = QtCore.QByteArray()

        if buf:
            buf = QtCore.qUncompress(buf)
            renderer.load(buf)
        else:
            # If no buffer is available, load from file
            if (mode, state) in self._svgFiles:
                svgFile = self._svgFiles[(mode, state)]
                renderer.load(self._replace_in_stream(svgFile))
            elif self.default_key in self._svgFiles:
                svgFile = self._svgFiles[self.default_key]
                if mode == QtGui.QIcon.Disabled:
                    renderer.load(self._replace_in_stream(svgFile, 'disabled'))
                else:
                    renderer.load(self._replace_in_stream(svgFile))

    def paint(self, painter, rect, mode, state):
        painter.drawPixmap(rect, self.pixmap(rect.size(), mode, state))

    def actualSize(self, size, mode, state):
        if (mode, state) in self._addedPixmaps:
            pm = self._addedPixmaps[(mode, state)]
            if pm and pm.size() == size:
                return size

        pm = self.pixmap(size, mode, state)
        if pm is None:
            return QtCore.QSize()
        return pm.size()

    def pixmap(self, size, mode, state):
        # Check if the palette has changed (if so invalidate cache)
        if self._palette_key != QtWidgets.QApplication.palette().cacheKey():
            QtGui.QPixmapCache.clear()
            self._set_replacements_from_palette()
        pmckey = self._make_cache_key(size, mode, state)
        pm = QtGui.QPixmapCache.find(pmckey)
        if pm:
            return pm

        if (mode, state) in self._addedPixmaps:
            pm = self._addedPixmaps[(mode, state)]
            if pm is not None and pm.size() == size:
                return pm

        renderer = QtSvg.QSvgRenderer()
        self._loadDataForModeAndState(renderer, mode, state)
        if not renderer.isValid():
            return QtGui.QPixmap()

        actualSize = renderer.defaultSize()
        if actualSize is not None:
            actualSize.scale(size, QtCore.Qt.KeepAspectRatio)

        img = QtGui.QImage(actualSize, QtGui.QImage.Format_ARGB32_Premultiplied)
        img.fill(0x00000000)
        p = QtGui.QPainter(img)
        renderer.render(p)
        p.end()
        pm = QtGui.QPixmap.fromImage(img)
        opt = QtWidgets.QStyleOption()
        opt.palette = QtWidgets.QApplication.palette()
        if self.use_qt_disabled or mode != QtGui.QIcon.Disabled:
            generated = QtWidgets.QApplication.style().generatedIconPixmap(mode, pm, opt)
            if generated is not None:
                pm = generated

        if pm is not None:
            QtGui.QPixmapCache.insert(pmckey, pm)

        return pm

    def addPixmap(self, pixmap, mode, state):
        self._addedPixmaps[(mode, state)] = pixmap

    def addFile(self, fileName, size, mode, state):
        if fileName:
            abs = fileName
            if fileName[0] != ':':
                abs = QtCore.QFileInfo(fileName).absoluteFilePath()
            al = abs.lower()
            if (al.endswith(".svg") or al.endswith(".svgz") or
                    al.endswith(".svg.gz")):
                renderer = QtSvg.QSvgRenderer(abs)
                if renderer.isValid():
                    self._svgFiles[(mode, state)] = abs
            else:
                pm = QtGui.QPixmap(abs)
                if pm is not None:
                    self.addPixmap(pm, mode, state)

    def key(self):
        return "svg"

    def clone(self):
        return SmartColorSVGIconEngine(other=self)

    def read(self, ds_in):

        self._svgBuffers = {}    # QHash<int, QByteArray>

        if ds_in.version() >= QtCore.QDataStream.Qt_4_4:
            nfiles = ds_in.readInt()
            fileNames = {}
            for i in range(nfiles):
                fileNames[i] = ds_in.readString()
            isCompressed = ds_in.readBool()
            key = ds_in.readInt()
            self._svgBuffers[key] = QtCore.QByteArray()
            ds_in >> self._svgBuffers[key]
            if not isCompressed:
                for key, v in self._svgBuffers.items():
                    self._svgBuffers[key] = QtCore.qCompress(v)
            hasAddedPixmaps = ds_in.readInt()
            if hasAddedPixmaps:
                npixmaps = ds_in.readInt()
                self._addedPixmaps = {}
                for i in range(npixmaps):
                    pm = QtGui.QPixmap()
                    ds_in >> pm
                    self._addedPixmaps[i] = pm
        else:
            pixmap = QtGui.QPixmap()
            data = QtCore.QByteArray()

            ds_in >> data
            if not data.isEmpty():
                data = QtCore.qUncompress(data)
                if not data.isEmpty():
                    self._svgBuffers[self.default_key] = data
            num_entries = ds_in.readInt()
            for i in range(num_entries):
                if ds_in.atEnd():
                    return False
                ds_in >> pixmap
                mode = ds_in.readUInt32()
                state = ds_in.readUInt32()
                # The pm list written by 4.3 is buggy and/or useless, so ignore
                # self._addPixmap(pixmap, QIcon.Mode(mode), QIcon.State(state))

        return True

    def write(self, ds_out):
        if out.version() >= QtCore.QDataStream.Qt_4_4:
            isCompressed = 1
            if self._svgBuffers:
                svgBuffers = self._svgBuffers
            else:
                svgBuffers = {}     # QHash<int, QByteArray>
            for key, v in self._svgFiles.items():
                buf = QtCore.QByteArray()
                f = QtCore.QFile(v)
                if f.open(QtCore.QIODevice.ReadOnly):
                    buf = f.readAll()
                buf = QtCore.qCompress(buf)
                svgBuffers[key] = buf
            out << self._svgFiles << isCompressed << svgBuffers
            if self._addedPixmaps:
                out << 1 << self._addedPixmaps
            else:
                out << 0
        else:
            buf = QtCore.QByteArray()
            if self._svgBuffers:
                buf = self._svgBuffers[self.default_key]
            if buf.isEmpty():
                svgFile = self._svgFiles[self.default_key]
                if not svgFile.isEmpty():
                    f = QtCore.QFile(svgFile)
                    if f.open(QtCore.QIODevice.ReadOnly):
                        buf = f.readAll()
            buf = QtCore.qCompress(buf)
            out << buf
            # 4.3 has buggy handling of added pixmaps, so don't write any
            out << 0
        return True



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Dec 12 23:44:01 2014

@author: Vidar Tonaas Fauske
"""


from hyperspyui.plugins.plugin import Plugin

from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QLineEdit, QLabel

from hyperspy.learn.mva import LearningResults
from hyperspyui.util import win2sig, fig2win, Namespace
from hyperspyui.threaded import ProgressThreaded, ProcessCanceled
from hyperspyui.widgets.extendedqwidgets import ExToolWindow


def tr(text):
    return QtCore.QCoreApplication.translate("MVA", text)


def align_yaxis(ax1, v1, ax2, v2):
    """adjust ax2 ylimit so that v2 in ax2 is aligned to v1 in ax1"""
    _, y1 = ax1.transData.transform((0, v1))
    _, y2 = ax2.transData.transform((0, v2))
    if y2 > y1:
        ratio = y1 / y2
    else:
        ratio = y2 / y1
    inv = ax2.transData.inverted()
    _, dy = inv.transform((0, 0)) - inv.transform((0, y1 - y2))
    miny2, maxy2 = ax2.get_ylim()
    ax2.set_ylim((miny2 + dy) / ratio, (maxy2 + dy) / ratio)


def make_advanced_dialog(ui, algorithms=None):
    diag = ExToolWindow(ui)

    diag.setWindowTitle("Decomposition parameters")

    vbox = QtWidgets.QVBoxLayout()
    if algorithms:
        lbl_algo = QLabel(tr("Choose algorithm:"))
        cbo_algo = QLineEdit.QComboBox()
        cbo_algo.addItems(algorithms)

        vbox.addWidget(lbl_algo)
        vbox.addWidget(cbo_algo)
    else:
        lbl_comp = QLabel(tr(
            "Enter a comma-separated list of component numbers to use for "
            "the model:"))
        txt_comp = QLineEdit()
        vbox.addWidget(lbl_comp)
        vbox.addWidget(txt_comp)

    btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                            QtCore.Qt.Horizontal)
    btns.accepted.connect(diag.accept)
    btns.rejected.connect(diag.reject)
    vbox.addWidget(btns)

    diag.setLayout(vbox)

    diag.algorithm = lambda: cbo_algo.currentText()
    diag.components = lambda: [int(s) for s in txt_comp.text().split(',')]
    return diag


class MVA_Plugin(Plugin):

    """
    Implements MVA decomposition utilities.
    """
    name = 'MVA'    # Used for settings groups etc

    coc_values = {'convert': tr("Convert"),
                  'copy': tr("Copy")}

    # ----------- Plugin interface -----------
    def create_actions(self):
        self.settings.set_default('convert_or_copy', None)
        self.settings.set_enum_hint('convert_or_copy',
                                    self.coc_values.keys())
        self.add_action('plot_decomposition_results',
                        tr("Decompose"),
                        self.plot_decomposition_results,
                        icon='pca.svg',
                        tip=tr("Decompose signal using Principle Component "
                               "analysis"),
                        selection_callback=self.selection_rules)
        self.add_action('pca', tr("Decomposition model"), self.pca,
                        icon='pca.svg',
                        tip=tr("Create a Principal Component Analysis "
                               "decomposition model"),
                        selection_callback=self.selection_rules)
        self.add_action('bss', tr("BSS"), self.bss,
                        icon='bss.svg',
                        tip=tr("Run Blind Source Separation"),
                        selection_callback=self.selection_rules)
        self.add_action('bss_model', tr("BSS model"), self.bss_model,
                        icon='bss.svg',
                        tip=tr("Create a Blind Source Separation "
                               "decomposition model"),
                        selection_callback=self.selection_rules)
        self.add_action('clear', tr("Clear"), self.clear,
                        tip=tr("Clear decomposition cache"),
                        selection_callback=self.selection_rules)

    def create_menu(self):
        self.add_menuitem('Decomposition',
                          self.ui.actions['plot_decomposition_results'])
        self.add_menuitem('Decomposition', self.ui.actions['pca'])
        self.add_menuitem('Decomposition', self.ui.actions['bss'])
        self.add_menuitem('Decomposition', self.ui.actions['bss_model'])
        self.add_menuitem('Decomposition', self.ui.actions['clear'])

    def create_toolbars(self):
        self.add_toolbar_button("Decomposition", self.ui.actions['pca'])
        self.add_toolbar_button("Decomposition", self.ui.actions['bss'])

    def selection_rules(self, win, action):
        """
        Callback to determine if action is valid for the passed window.
        """
        s = win2sig(win, self.ui.signals)
        if s is None or s.signal.data.ndim <= 1:
            action.setEnabled(False)
        else:
            action.setEnabled(True)

    # ------------ Action implementations --------------

    def _get_signal(self, signal):
        """
        Get a valid signal. If the signal is none, it ues the currently
        selected one. If the signal type is not float, it either converts it,
        or gets a copy of the correct type, depending on the 'convert_copy'
        setting.
        """
        if signal is None:
            signal = self.ui.get_selected_wrapper()
        s = signal.signal

        if s.data.dtype.char not in ['e', 'f', 'd']:  # If not float
            cc = self.settings.get_or_prompt(
                'convert_or_copy',
                [kv for kv in self.coc_values.items()],
                title=tr("Convert or copy"),
                descr=tr(
                    "Signal data has the wrong data type (float needed)." +
                    "Would you like to convert the current signal, or " +
                    "perform the decomposition on a copy?"))
            if cc is None:
                # User canceled
                raise ProcessCanceled()
            if cc == 'copy':
                s = s.deepcopy()
                s.metadata.General.title = signal.name + "[float]"
                s.plot()
            s.change_dtype(float)
        return s, signal

    def _do_decomposition(self, s, force=False, algorithm=None):
        """
        Makes sure we have decomposition results. If results already are
        available, it will only recalculate if the `force` parameter is True.
        """
        if algorithm:
            s.decomposition(algorithm=algorithm)
        elif force or s.learning_results.explained_variance_ratio is None:
            s.decomposition()
        return s

    def _do_bss(self, s, n_components, algorithm=None):
        """
        Makes sure we have BSS results. If results already are available, it
        will only recalculate if the `force` parameter is True.
        """
        if algorithm:
            s.blind_source_separation(n_components, algorithm=algorithm)
        else:
            s.blind_source_separation(n_components)

    def get_bss_results(self, signal):
        factors = signal.get_bss_factors()
        loadings = signal.get_bss_loadings()
        factors.axes_manager._axes[0] = loadings.axes_manager._axes[0]
        return loadings, factors

    def _record(self, autosig, model, signal, n_components):
        if autosig:
            self.record_code(r"<p>.{0}(n_components={1})".format(
                             model, n_components))
        else:
            self.record_code(r"<p>.{0}({1}, n_components={2})".format(
                             model, signal, n_components))

    def _decompose_threaded(self, callback, label, signal=None,
                            algorithm=None, ns=None):
        if ns is None:
            ns = Namespace()
            ns.autosig = signal is None
            ns.s, signal = self._get_signal(signal)

        def do_threaded():
            ns.s = self._do_decomposition(ns.s, algorithm=algorithm)

        def on_error(message=None):
            em = QtWidgets.QErrorMessage(self.ui)
            msg = tr("An error occurred during decomposition")
            if message:
                msg += ":\n" + message
            em.setWindowTitle(tr("Decomposition error"))
            em.showMessage(msg)

        t = ProgressThreaded(self.ui, do_threaded, lambda: callback(ns),
                             label=label)
        t.worker.error[str].connect(on_error)
        t.run()

    def _perform_model(self, ns, n_components):
        # Num comp. picked, get model, wrap new signal and plot
        if ns.model == 'pca':
            sc = ns.s.get_decomposition_model(n_components)
            sc.metadata.General.title = ns.signal.name + "[PCA-model]"
            sc.plot()
        elif ns.model == 'bss' or ns.model.startswith('bss.'):
            if ns.model.startswith('bss.'):
                algorithm = ns.model[len('bss.'):]
                self._do_bss(ns.s, n_components, algorithm=algorithm)
            else:
                self._do_bss(ns.s, n_components)
            f, o = self.get_bss_results(ns.s)
            o.metadata.add_dictionary(ns.s.metadata.as_dictionary())
            f.metadata.General.title = ns.signal.name + "[BSS-Factors]"
            o.metadata.General.title = ns.signal.name + "[BSS-Loadings]"
            f.plot()
            o.plot()
        elif ns.model == 'bss_model':
            # Here we have to assume the user has actually performed the BSS
            # decomposition first!
            sc = ns.s.get_bss_model(n_components)
            sc.metadata.General.title = ns.signal.name + "[BSS-model]"
            sc.plot()
        if not ns.recorded:
            self._record(ns.autosig, ns.model, ns.signal, n_components)

    def _show_scree(self, ns, callback):
        ax = ns.s.plot_explained_variance_ratio()

        # Clean up plot and present, allow user to select components
        # by picker
        ax.set_title("")
        scree = ax.get_figure().canvas
        scree.draw()
        scree.setWindowTitle("Pick number of components")

        def clicked(event):
            n_components = int(round(event.xdata))
            # Close scree plot
            w = fig2win(scree.figure, self.ui.figures)
            w.close()
            callback(ns, n_components)
        scree.mpl_connect('button_press_event', clicked)

    def do_after_scree(self, model, signal=None, n_components=None):
        """
        Performs decomposition, then plots the scree for the user to select
        the number of components to use for a decomposition model. The
        selection is made by clicking on the scree, which closes the scree
        and creates the model.
        """
        ns = Namespace()
        ns.autosig = signal is None
        ns.model = model
        ns.s, ns.signal = self._get_signal(signal)
        if n_components is not None:
            self._record(ns.autosig, ns.model, ns.signal, n_components)
            ns.recorded = True
        else:
            ns.recorded = False

        def on_complete(ns):
            if n_components is None:
                self._show_scree(ns, self._perform_model)
            else:
                self._perform_model(ns, n_components)

        self._decompose_threaded(on_complete, "Performing %s" % model.upper(),
                                 n_components, ns=ns)

    def plot_decomposition_results(self, signal=None, advanced=False):
        """
        Performs decomposition if necessary, then plots the decomposition
        results according to the hyperspy implementation.
        """
        def on_complete(ns):
            ns.s.plot_decomposition_results()
            # Somewhat speculative workaround to HSPY not adding metadata
            sd = self.ui.hspy_signals[-1]
            sd.metadata.add_dictionary(ns.s.metadata.as_dictionary())

        if advanced:
            diag = make_advanced_dialog(
                self.ui, ['svd', 'fast_svd', 'mlpca', 'fast_mlpca', 'nmf',
                          'sparse_pca', 'mini_batch_sparse_pca'])
            dr = diag.exec_()
            if dr == QDialog.Accepted:
                self._decompose_threaded(
                    on_complete, "Decomposing signal",
                    algorithm=diag.algorithm())
        else:
            self._decompose_threaded(on_complete, "Decomposing signal")

    def pca(self, signal=None, n_components=None, advanced=False):
        """
        Performs decomposition, then plots the scree for the user to select
        the number of components to use for a decomposition model. The
        selection is made by clicking on the scree, which closes the scree
        and creates the model.
        """
        if advanced:
            diag = make_advanced_dialog(self.ui)
            dr = diag.exec_()
            if dr == QDialog.Accepted:
                self.do_after_scree(
                    'pca', signal, n_components=diag.components())
        else:
            self.do_after_scree('pca', signal, n_components)

    def bss(self, signal=None, n_components=None, advanced=False):
        """
        Performs decomposition if neccessary, then plots the scree for the user
        to select the number of components to use for a blind source
        separation. The selection is made by clicking on the scree, which
        closes the scree and creates the model.
        """
        if advanced:
            diag = make_advanced_dialog(
                self.ui, ['orthomax', 'sklearn_fastica', 'FastICA', 'JADE',
                          'CuBICA', 'TDSEP'])
            dr = diag.exec_()
            if dr == QDialog.Accepted:
                model = 'bss.' + diag.algorithm()
                self.do_after_scree(model, signal, n_components)
        else:
            self.do_after_scree('bss', signal, n_components)

    def bss_model(self, signal=None, n_components=None, advanced=False):
        """
        Performs decomposition if neccessary, then plots the scree for the user
        to select the number of components to use for a blind source
        separation model. The selection is made by clicking on the scree, which
        closes the scree and creates the model.
        """
        if advanced:
            diag = make_advanced_dialog(self.ui)
            dr = diag.exec_()
            if dr == QDialog.Accepted:
                self.do_after_scree(
                    'bss_model', signal, n_components=diag.components())
        else:
            self.do_after_scree('bss_model', signal, n_components)

    def clear(self, signal=None):
        """
        Clears the learning results from the signal.
        """
        if signal is None:
            signal = self.ui.get_selected_signal()
        signal.learning_results = LearningResults()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from hyperspyui.plugins.plugin import Plugin
import numpy as np
from skimage.filters import gaussian
from hyperspyui.util import win2sig
from hyperspyui.widgets.extendedqwidgets import ExToolWindow

from qtpy import QtGui, QtCore


def tr(text):
    return QtCore.QCoreApplication.translate("GaussianFilter", text)


class GaussianFilter(Plugin):
    name = "Gaussian Filter"

    def __init__(self, ui):
        super(GaussianFilter, self).__init__(ui)
        self.settings.set_default('sigma', 1.0)

    def create_actions(self):
        self.add_action(
            self.name + '.show_dialog', self.name, self.show_dialog,
            icon="gaussian_bare.svg",
            tip="Apply a gaussian filter",
            selection_callback=self.selection_rules)

    def create_menu(self):
        self.add_menuitem('Image', self.ui.actions[self.name + '.show_dialog'])

    def create_toolbars(self):
        self.add_toolbar_button(
            'Image', self.ui.actions[
                self.name + '.show_dialog'])

    def selection_rules(self, win, action):
        """
        Callback to determine if filter is valid for the passed window.
        """
        s = win2sig(win, self.ui.signals)
        ok = False
        if s is not None:
            if win == s.signal_plot and \
                    s.signal.axes_manager.signal_dimension == 2:
                ok = True
        if ok:
            action.setEnabled(True)
        else:
            action.setEnabled(False)

    def gaussian(self, sigma, signal=None, out=None, record=True,
                 *args, **kwargs):
        """
        Apply a gaussian smoothing filter to an image signal.
        Uses py:func:`skimage.filters.gaussian` for the actual processing.

        Parameters
        ----------
        sigma : {float}
            Smoothing factor in units of pixels, i.e a value around 1 is
            a slight smoothing.
        signal : {Signal | None}
            Signal to operate on. If not, it will use the currently
            selected one
        out : {Signal | None}
            Output signal
        record : {bool}
            Whether the operation should be recorded or not.
        **kwargs : dict
            Other keyword arguments are passed to 
            py:func:`skimage.filters.gaussian`.
        """
        if signal is None:
            signal, axes, _ = self.ui.get_selected_plot()
            signal = signal.signal
            if isinstance(axes, str):
                axm = signal.axes_manager
                if axes.startswith("nav"):
                    axes = (axm._axes.index(axm.navigation_axes[0]),
                            axm._axes.index(axm.navigation_axes[1]))
                elif axes.startswith("sig"):
                    axes = (axm._axes.index(axm.signal_axes[0]),
                            axm._axes.index(axm.signal_axes[1]))

        s_out = out or signal.deepcopy()
        if out is None and not np.issubdtype(s_out.data.dtype, np.float):
            s_out.change_dtype(np.float)
        if signal.data.dtype is np.float:
            vmin, vmax = np.nanmin(signal.data), np.nanmax(signal.data)
        else:
            vmin, vmax = (None, None)
        for im_o, im_i in zip(s_out._iterate_signal(),
                              signal._iterate_signal()):
            if vmin is None:
                im_o[:] = gaussian(im_i, sigma, *args, **kwargs)
            else:
                im_o[:] = gaussian((im_i-vmin)/(vmax-vmin), sigma, *args, **kwargs)
        if vmin is not None:
            s_out.data[:] = (vmax-vmin)*s_out.data + vmin

        if out is None:
            if record:
                self.record_code(r"<p>.gaussian({0}, {1}, {2})".format(
                        sigma, args, kwargs))
            return s_out
        else:
            if record:
                self.record_code(
                    r"<p>.gaussian({0}, out={1}, {2}, {3})".format(
                        sigma, out, args, kwargs))
            if hasattr(out, 'events') and hasattr(out.events, 'data_changed'):
                out.events.data_changed.trigger(out)

    def on_dialog_accept(self, dialog):
        self.settings['sigma'] = dialog.num_sigma.value()

    def show_dialog(self):
        signal, space, _ = self.ui.get_selected_plot()
        if space != "signal":
            return
        dialog = GaussianFilterDialog(signal, self.ui, self)
        if 'sigma' in self.settings:
            dialog.num_sigma.setValue(float(self.settings['sigma']))
        dialog.accepted.connect(lambda: self.on_dialog_accept(dialog))
        self.open_dialog(dialog)


class GaussianFilterDialog(ExToolWindow):

    def __init__(self, signal, parent, plugin):
        super(GaussianFilterDialog, self).__init__(parent)
        self.ui = parent
        self.create_controls()
        self.accepted.connect(self.ok)
        self.rejected.connect(self.undo)
        self.signal = signal
        self.plugin = plugin
        self.new_out = None
        self._connected_updates = False
        self.setWindowTitle(tr("Gaussian Filter"))

        # TODO: TAG: Functionality check
        if not hasattr(signal.signal, 'events'):
            self.chk_preview.setVisible(False)

    def connect(self):
        self.opt_new.toggled.connect(self.close_new)
        self.num_sigma.valueChanged.connect(self.update)
        self.opt_new.toggled.connect(self.update)
        self.opt_replace.toggled.connect(self.update)

    def disconnect(self):
        self.opt_new.toggled.disconnect(self.close_new)
        self.num_sigma.valueChanged.disconnect(self.update)
        self.opt_new.toggled.disconnect(self.update)
        self.opt_replace.toggled.disconnect(self.update)

    def ok(self):
        # Draw figure if not already done
        # TODO: TAG: Functionality check
        if not hasattr(self.signal.signal, 'events') or \
                not self.chk_preview.isChecked():
            self.update()
        sigma = self.num_sigma.value()
        self.plugin.record_code(r"<p>.gaussian({0})".format(sigma))

    def undo(self):
        self.close_new()
        self.set_preview

    def close_new(self, value=False):
        if self.new_out is not None and not value:
            self.new_out.close()
            self.new_out = None

    def set_preview(self, value):
        if not hasattr(self.signal.signal, 'events'):
            return
        if value:
            self.connect()
            self.update()
        else:
            self.disconnect()
            self.close_new()

    def update(self):
        sigma = self.num_sigma.value()
        if self.opt_new.isChecked():
            if self.new_out is None:
                out = None
            else:
                out = self.new_out.signal
        elif self.opt_replace.isChecked():
            out = self.signal.signal
        else:
            return  # Indeterminate state, do nothing

        s = self.plugin.gaussian(sigma, self.signal.signal,
                                 record=False, out=out)

        if out is None:
            s.metadata.General.title = self.signal.name + "[GaussianFilter]"
            s.plot()
            if (self.chk_preview.isChecked() and self.opt_new.isChecked() and
                    self.new_out is None):
                self.new_out = self.ui.lut_signalwrapper[s]

    def create_controls(self):
        """
        Create UI controls.
        """
        vbox = QtGui.QVBoxLayout()

        form = QtGui.QFormLayout()
        self.num_sigma = QtGui.QDoubleSpinBox()
        self.num_sigma.setValue(1.0)
        self.num_sigma.setMinimum(0.0)
        self.num_sigma.setSingleStep(0.1)
        self.num_sigma.setMaximum(1e3)
        self.num_sigma.setDecimals(2)
        form.addRow(tr("Sigma:"), self.num_sigma)
        vbox.addLayout(form)

        self.chk_preview = QtGui.QCheckBox(tr("Preview"))
        self.chk_preview.setCheckable(True)
        self.chk_preview.setChecked(False)
        vbox.addWidget(self.chk_preview)

        self.chk_preview.toggled[bool].connect(self.set_preview)

        self.gbo_output = QtGui.QGroupBox(tr("Output"))
        self.opt_new = QtGui.QRadioButton(tr("New signal"))
        self.opt_replace = QtGui.QRadioButton(tr("In place"))
        self.opt_new.setChecked(True)
        gbo_vbox2 = QtGui.QVBoxLayout()
        gbo_vbox2.addWidget(self.opt_new)
        gbo_vbox2.addWidget(self.opt_replace)
        self.gbo_output.setLayout(gbo_vbox2)
        vbox.addWidget(self.gbo_output)

        self.btn_ok = QtGui.QPushButton(tr("&OK"))
        self.btn_ok.setDefault(True)
        self.btn_ok.clicked.connect(self.accept)
        self.btn_cancel = QtGui.QPushButton(tr("&Cancel"))
        self.btn_cancel.clicked.connect(self.reject)
        hbox = QtGui.QHBoxLayout()
        hbox.addWidget(self.btn_ok)
        hbox.addWidget(self.btn_cancel)
        vbox.addLayout(hbox)

        vbox.addStretch(1)
        self.setLayout(vbox)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Oct 31 14:22:53 2014

@author: Vidar Tonaas Fauske
"""

import os
from distutils.version import LooseVersion

import matplotlib
from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import FigureManagerBase
from matplotlib.figure import Figure
from qtpy import API, QtCore, QtGui, QtWidgets


# FigureCanvas definition
if LooseVersion(matplotlib.__version__) >= LooseVersion('2.1.0'):
    # qt4 and qt5 are the same in matplotlib >= 2.1.0
    from matplotlib.backends import backend_qt5agg
    FigureCanvas = backend_qt5agg.FigureCanvasQTAgg
elif LooseVersion(matplotlib.__version__) >= LooseVersion('2.0.0'):
    if API == 'pyqt5':
        from matplotlib.backends import backend_qt5agg
        FigureCanvas = backend_qt5agg.FigureCanvasQTAgg
    else:
        from matplotlib.backends import backend_qt4agg
        FigureCanvas = backend_qt4agg.FigureCanvasQTAgg
else:  # < 2.0.0
    from matplotlib.backends import backend_qt4agg
    FigureCanvas = backend_qt4agg.FigureCanvas


# Workaround issue with matplotlib interactive backend and ipython integration
# https://github.com/matplotlib/matplotlib/issues/14426
# issue introduced in https://github.com/matplotlib/matplotlib/pull/12637
# The dummy `FigureCanvas(Figure())` initialise ipython/matplotlib support when
# creating the dummy figure which makes everyone happy!
FigureCanvas(Figure())

# =================
# Event managers
# =================

_new_fig_cbs = {}
_destroy_cbs = {}


def connect_on_new_figure(callback, userdata=None):
    """
    Call to subscribe to new MPL figure events. 'callback' is called on the
    event, with the figure as it's first parameter, and 'userdata' as it's
    second parameter if it is not None. If it's None, only one parameter is
    passed.
    """
    global _new_fig_cbs
    _new_fig_cbs[callback] = userdata


def disconnect_on_new_figure(callback):
    """
    Disconnect callback from subscription.
    """
    global _new_fig_cbs
    if callback in _new_fig_cbs:
        _new_fig_cbs.pop(callback)


def _on_new_figure(figure):
    """
    New MPL figure event, calls subscribers.
    'figure' parameter is of the type FigureWindow defined below
    """
    global _new_fig_cbs
    for callback, userdata in _new_fig_cbs.items():
        try:
            callback(figure, userdata)
        except TypeError:
            callback(figure)


def connect_on_destroy(callback, userdata=None):
    """
    Call to subscribe to destroying MPL figure events. 'callback' is called on
    the event, with the figure as it's first parameter, and 'userdata' as it's
    second parameter if it is not None. If it's None, only one parameter is
    passed.
    """
    global _destroy_cbs
    _destroy_cbs[callback] = userdata


def disconnect_on_destroy(callback):
    """
    Disconnect callback from subscription.
    """
    global _destroy_cbs
    if callback in _destroy_cbs:
        _destroy_cbs.pop(callback)


def _on_destroy(figure):
    """
    Destroying MPL figure event, calls subscribers.
    'figure' parameter is of the type FigureWindow defined below
    """
    global _destroy_cbs
    for callback, userdata in _destroy_cbs.items():
        try:
            callback(figure, userdata)
        except TypeError:
            callback(figure)


def new_figure_manager(num, *args, **kwargs):
    """
    Create a new figure manager instance. MPL backend function.
    """
    FigureClass = kwargs.pop(
        'FigureClass', matplotlib.figure.Figure)
    thisFig = FigureClass(*args, **kwargs)
    return new_figure_manager_given_figure(num, thisFig)


def new_figure_manager_given_figure(num, figure):
    """
    Create a new figure manager instance for the given figure. MPL backend
    function.
    """
    canvas = FigureCanvas(figure)
    manager = FigureManagerMdi(canvas, num)
    return manager


class FigureWindow(QtWidgets.QMdiSubWindow):

    """
    A basic MDI sub-window, but with a closing signal, and an activate QAction,
    which allows for switching between all FigureWindows (e.g. by a
    Windows-menu). An exclusive, static action group makes sure only one
    window can be active at the time. If you want to split the windows into
    different groups that can be treated separately, you will need to create
    your own QtWidgets.QActionGroups.
    """
    closing = QtCore.Signal()

    activeFigureActionGroup = QtWidgets.QActionGroup(None)
    activeFigureActionGroup.setExclusive(True)

    def __init__(self, *args, **kwargs):
        super(FigureWindow, self).__init__(*args, **kwargs)
        self._activate_action = None
        if API != 'pyside':
            self.windowStateChanged.connect(self._windowStateChanged)

    def closeEvent(self, event):
        self.closing.emit()
        super(FigureWindow, self).closeEvent(event)

    def activateAction(self):
        """
        Returns a QAction that will activate the window with setActiveSubWindow
        as long as it has an mdiArea set. If not, it will use activateWindow to
        try to make it the active window.
        """
        if self._activate_action is not None:
            return self._activate_action
        self._activate_action = QtWidgets.QAction(self.windowTitle(), self)
        self._activate_action.setCheckable(True)
        self._activate_action.triggered.connect(self._activate_triggered)
        self.activeFigureActionGroup.addAction(self._activate_action)
        return self._activate_action

    def setWindowTitle(self, title):
        # Overridden to keep action text updated
        super(FigureWindow, self).setWindowTitle(title)
        if self._activate_action is not None:
            self._activate_action.setText(title)

    def _windowStateChanged(self, oldState, newState):
        # Event for tracking if we're active or not
        isactive = newState & QtCore.Qt.WindowActive
        if isactive == oldState & QtCore.Qt.WindowActive:
            return  # Another window state changed, e.g. activation
        self._activate_action.setChecked(isactive)

    def _activate_triggered(self, checked=True):
        # Activate action triggered, make window active
        if self.mdiArea():
            self.mdiArea().setActiveSubWindow(self)
        else:
            self.activateWindow()   # If not subwindow

        if self.isMinimized():
            self.showNormal()   # Restore minimized window
        if not checked:
            # User unchecked, which makes no sense, recheck
            self._activate_action.setChecked(True)


class FigureManagerMdi(FigureManagerBase):

    """
    Public attributes

    canvas      : The FigureCanvas instance
    num         : The Figure number
    toolbar     : The qt.QToolBar
    window      : The FigureWindow

    Our MPL figure manager class. Much of the code is copied from MPL Qt4
    backend, but adapted for our needs.
    """

    def __init__(self, canvas, num):
        FigureManagerBase.__init__(self, canvas, num)
        self.canvas = canvas
        self.window = FigureWindow()
        self.window.closing.connect(canvas.close_event)
        self.window.closing.connect(self._widgetclosed)

        self.window.setWindowTitle("Figure %d" % num)
        image = os.path.join(matplotlib.rcParams['datapath'],
                             'images', 'matplotlib.png')
        self.window.setWindowIcon(QtGui.QIcon(image))

        # Give the keyboard focus to the figure instead of the
        # manager; StrongFocus accepts both tab and click to focus and
        # will enable the canvas to process event w/o clicking.
        # ClickFocus only takes the focus is the window has been
        # clicked
        # on. http://qt-project.org/doc/qt-4.8/qt.html#FocusPolicy-enum or
        # http://doc.qt.digia.com/qt/qt.html#FocusPolicy-enum
        self.canvas.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.canvas.setFocus()

        self.window._destroying = False

        # No toolbar in figure currently
        self.toolbar = None
        tbs_height = 0

        # resize the main window so it will display the canvas with the
        # requested size:
        cs = canvas.sizeHint()
        self._status_and_tool_height = tbs_height
        height = cs.height() + self._status_and_tool_height
        self.window.resize(cs.width(), height)

        self.window.setWidget(self.canvas)

        _on_new_figure(self.window)

        if matplotlib.is_interactive():
            self.window.show()

        def notify_axes_change(fig):
            # This will be called whenever the current axes is changed
            if self.toolbar is not None:
                self.toolbar.update()
        self.canvas.figure.add_axobserver(notify_axes_change)

    @QtCore.Slot()
    def _show_message(self, s):
        # Fixes a PySide segfault.
        print("Trying to show: " + s)
#        self.window.statusBar().showMessage(s)

    def full_screen_toggle(self):
        if self.window.isFullScreen():
            self.window.showNormal()
        else:
            self.window.showFullScreen()

    def _widgetclosed(self):
        if self.window is None or self.window._destroying:
            return
        try:
            Gcf.destroy(self.num)
        except AttributeError:
            pass
            # It seems that when the python session is killed,
            # Gcf can get destroyed before the Gcf.destroy
            # line is run, leading to a useless AttributeError.

    def resize(self, width, height):
        'set the canvas size in pixels'
        self.window.resize(width, height + self._status_and_tool_height)

    def show(self):
        self.window.show()

    def destroy(self, *args):
        _on_destroy(self.window)
        # check for qApp first, as PySide deletes it in its atexit handler
        if QtWidgets.QApplication.instance() is None:
            return
        if self.window._destroying:
            return
        self.window._destroying = True
        self.window.destroyed.connect(self._widgetclosed)

        if self.toolbar:
            self.toolbar.destroy()
        self.window.close()
        self.window = None
        # self.canvas.figure = None  # Causes exceptions in delayed drawing
        self.canvas = None
        self.toolbar = None

    def get_window_title(self):
        return str(self.window.windowTitle())

    def set_window_title(self, title):
        self.window.setWindowTitle(title)


# Definition for MPL backend:
FigureManager = FigureManagerMdi



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon Oct 27 23:09:55 2014

@author: Vidar Tonaas Fauske
"""


from qtpy import QtWidgets
from qtpy.QtCore import Qt

from hyperspyui.bindinglist import BindingList


class SignalList(QtWidgets.QListWidget):

    def __init__(self, items=None, parent=None, multiselect=True):
        super(SignalList, self).__init__(parent)
        self.multiselect = multiselect
        self._bound_blists = []

        if items is not None:
            self.addItems(items)
            if isinstance(items, BindingList):
                self.bind(items)
        self.destroyed.connect(self._on_destroy)

    @property
    def multiselect(self):
        return self._multiselect

    @multiselect.setter
    def multiselect(self, value):
        self._multiselect = value
        if value:
            self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        else:
            self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

    def bind(self, blist):
        blist.add_target(self)
        self._bound_blists.append(blist)

    def unbind(self, blist):
        blist.remove_target(self)
        self._bound_blists.remove(blist)

    def _on_destroy(self):
        for b in reversed(self._bound_blists):
            self.unbind(b)

    def addItem(self, object):
        item = QtWidgets.QListWidgetItem(object.name, self)
        item.setData(Qt.UserRole, object)
        super(SignalList, self).addItem(item)

    def addItems(self, items):
        for i in items:
            self.addItem(i)

    def insertItem(self, index, object):
        item = QtWidgets.QListWidgetItem(object.name, self)
        item.setData(Qt.UserRole, object)
        super(SignalList, self).insertItem(index, item)

    def signal(self, index):
        if isinstance(index, int):
            return self.item(index).data(Qt.UserRole)
        elif isinstance(index, QtWidgets.QListWidgetItem):
            return index.data(Qt.UserRole)

    def get_selected(self):
        selections = self.selectedItems()
        sigs = [self.signal(i) for i in selections]
        if self.multiselect:
            return sigs
        else:
            if len(sigs) > 0:
                return sigs[0]
            else:
                return None

    def __getitem__(self, key):
        if isinstance(key, QtWidgets.QListWidgetItem):
            return key.data(Qt.UserRole)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Thu Nov 27 03:01:19 2014

@author: Vidar Tonaas Fauske
"""


import sys
import json
import os
import logging
from qtpy import QtGui, QtCore, QtNetwork, API, QtWidgets

_logger = logging.getLogger(__name__)


class SingleApplication(QtWidgets.QApplication):
    messageAvailable = QtCore.Signal(object)

    def __init__(self, argv, key):
        QtWidgets.QApplication.__init__(self, argv)
        self._memory = QtCore.QSharedMemory(self)
        self._memory.setKey(key)
        if self._memory.attach():
            self._running = True
        else:
            self._running = False
            if not self._memory.create(1):
                raise RuntimeError(self._memory.errorString())

        # Set correct logo
        base_path = os.path.abspath(os.path.dirname(__file__))
        icon_path = os.path.join(base_path, 'images', 'hyperspy.svg')
        QtWidgets.QApplication.setWindowIcon(QtGui.QIcon(icon_path))
        self.aboutToQuit.connect(self.detach_shared_memory)

    def isRunning(self):
        return self._running

    def detach_shared_memory(self):
        _logger.debug('Detaching shared memory.')
        if not self._memory.detach():
            _logger.debug('Shared memory could not be detached properly.')


class SingleApplicationWithMessaging(SingleApplication):

    def __init__(self, argv, key):
        SingleApplication.__init__(self, argv, key)
        self._key = key
        self._timeout = 1000
        self._server = QtNetwork.QLocalServer(self)
        if not self.isRunning():
            self._server.newConnection.connect(self.handleMessage)
            self._server.listen(self._key)

    def handleMessage(self):
        socket = self._server.nextPendingConnection()
        if socket.waitForReadyRead(self._timeout):
            self.messageAvailable.emit(
                socket.readAll().data().decode('utf-8'))
            socket.disconnectFromServer()
        else:
            QtCore.qDebug(socket.errorString())

    def sendMessage(self, message):
        if self.isRunning():
            socket = QtNetwork.QLocalSocket(self)
            socket.connectToServer(self._key, QtCore.QIODevice.WriteOnly)
            if not socket.waitForConnected(self._timeout):
                _logger.error(socket.errorString())
                return False
            if not isinstance(message, bytes):
                message = message.encode('utf-8')
            socket.write(message)
            if not socket.waitForBytesWritten(self._timeout):
                _logger.error(socket.errorString())
                return False
            socket.disconnectFromServer()
            return True
        return False


def get_app(key):
    # send commandline args as message
    if "pyqt" in API:
        if len(sys.argv) > 1:
            app = SingleApplicationWithMessaging(sys.argv, key)
            if app.isRunning():
                msg = json.dumps(sys.argv[1:])
                app.sendMessage(msg)
                _logger.debug('An existing instance of HyperSpyUI is running, '
                              'sending arguments to it.')
                sys.exit(1)     # An instance is already running
        else:
            app = SingleApplicationWithMessaging(sys.argv, key)
            if app.isRunning():
                _logger.debug('An existing instance of HyperSpyUI is running, '
                              'bringing it to the front.')
                sys.exit(1)     # An instance is already running
    elif API == 'pyside':
        from siding.singleinstance import QSingleApplication
        app = QSingleApplication(sys.argv)
        msg = json.dumps(sys.argv[1:])
        app.ensure_single(message=msg)
    else:
        app = QtWidgets.QApplication(sys.argv)
    return app



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.

from hyperspyui.plugins.plugin import Plugin
import numpy as np
from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QDialogButtonBox


from hyperspyui.tools import SelectionTool
from hyperspyui.util import SignalTypeFilter
from hyperspyui.widgets.extendedqwidgets import ExToolWindow

from hyperspy.roi import BaseInteractiveROI
import hyperspy.signals


class AlignPlugin(Plugin):
    name = "Align"

    def __init__(self, main_window):
        super(AlignPlugin, self).__init__(main_window)
        self.settings.set_default('sub_pixel_factor', 20)
        self.settings.set_default('1d_smooth_amount', 50)
        self.settings.set_default('2d_smooth_amount', 0.0)
        self.settings.set_default('sobel_2D', True)
        self.settings.set_default('median_2D', True)
        self.settings.set_default('hanning_2D', True)
        self.settings.set_default('alignment_reference', 'current')
        self.settings.set_default('expand', True)
        self.settings.set_default('crop', True)
        self.settings.set_default('plot', False)

    def create_actions(self):
        self.add_action('manual_align', "Manual align",
                        self.manual_align,
                        icon='align_manual.svg',
                        tip="Interactively align the signal",
                        selection_callback=SignalTypeFilter(
                            hyperspy.signals.Signal2D, self.ui))

    def create_menu(self):
        self.add_menuitem("Signal", self.ui.actions['manual_align'])

    def create_toolbars(self):
        self.add_toolbar_button("Signal", self.ui.actions['manual_align'])

    def create_tools(self):
        tools = []

        # XD tool
        self.tool_XD = SelectionTool(
            name='Align tool', icon="align2d.svg", category="Signal",
            description="Align images across the stack")
        self.tool_XD.accepted[BaseInteractiveROI].connect(
            self.align_XD)
        self.tool_XD.valid_dimensions = (1, 2)
        tools.append(self.tool_XD)

        # Vertical 2D align
        self.tool_vertical = SelectionTool(
            name='Align vertical tool', icon="align_vertical.svg",
            category="Signal",
            description="Align an image feature vertically across the stack")
        self.tool_vertical.accepted[BaseInteractiveROI].connect(
            self.align_vertical)
        self.tool_vertical.valid_dimensions = (2,)
        tools.append(self.tool_vertical)

        # Vertical 2D align
        self.tool_horizontal = SelectionTool(
            name='Align horizontal tool', icon="align_horizontal.svg",
            category="Signal",
            description="Align an image feature horizontally across the stack")
        self.tool_horizontal.accepted[BaseInteractiveROI].connect(
            self.align_horizontal)
        self.tool_horizontal.valid_dimensions = (2,)
        tools.append(self.tool_horizontal)

        for t in tools:
            t.cancel_on_accept = True
            self.add_tool(t, self.ui.select_signal)

    def manual_align(self, signal=None):
        signal = self._get_signal(signal)
        if signal is None:
            return
        diag = ManualAlignDialog(signal, self.ui)
        diag.show()

    @staticmethod
    def _smooth(y, box_pts):
        box = np.ones(box_pts) / box_pts
        y_smooth = np.convolve(y, box, mode='valid')
        return y_smooth

    def _get_signal(self, signal):
        if signal is None:
            return self.ui.get_selected_signal()
        return signal

    def align_XD(self, roi, signal=None):
        signal = self._get_signal(signal)
        if signal is None:
            return
        if signal.axes_manager.signal_dimension != roi.ndim:
            return
        if roi.ndim == 1:
            return self.align_1D(roi, signal)
        elif roi.ndim == 2:
            return self.align_2D(roi, signal)
        else:
            raise ValueError("Cannot align a signal of %d dimensions" %
                             roi.ndim)

    def align_1D(self, roi, signal=None):
        signal = self._get_signal(signal)
        if signal is None:
            return
        shifts = signal.estimate_shift1D(
            reference='current',
            roi=(roi.left, roi.right),
            show_progressbar=True)
        s_aligned = signal.deepcopy()
        s_aligned.align1D(shifts=shifts, expand=True)
        s_aligned.plot()
        self.record_code("signal = ui.get_selected_signal()")
        self.record_code("s_aligned = signal.deepcopy()")
        self.record_code("s_aligned.align1D(reference='current', "
                         "roi=(%f, %f), show_progressbar=True, expand=True)" %
                         (roi.left, roi.right))
        return s_aligned

    def align_2D(self, roi, signal=None):
        signal = self._get_signal(signal)
        if signal is None:
            return
        s = signal
        sobel = self.settings['sobel_2D', bool]
        hanning = self.settings['hanning_2D', bool]
        median = self.settings['median_2D', bool]
        sub_pixel_factor = self.settings['sub_pixel_factor', float]
        plot = self.settings['plot', bool]
        if plot:
            plot = 'reuse'
        ref = self.settings['alignment_reference'].lower()
        if not ref:
            ref = 'current'
        expand = self.settings['expand', bool]
        crop = self.settings['crop', bool]
        gauss = self.settings['2d_smooth_amount', float]
        if gauss > 0.0 and 'Gaussian Filter' in self.ui.plugins:
            p = self.ui.plugins['Gaussian Filter']
            s = p.gaussian(sigma=gauss, signal=signal, record=False)
            s.axes_manager.indices = signal.axes_manager.indices
        record_string = (
            "reference={0}, sobel={1}, hanning={2}, medfilter={3},"
            "roi=({4}, {5}, {6}, {7}), plot={8},"
            "show_progressbar=True").format(
                ref, sobel, hanning, median, roi.left, roi.right, roi.top,
                roi.bottom, plot)
        try:
            shifts = s.estimate_shift2D(
                reference=ref,
                roi=(roi.left, roi.right, roi.top, roi.bottom),
                sobel=sobel, hanning=hanning, medfilter=median,
                sub_pixel_factor=sub_pixel_factor,
                plot=plot,
                show_progressbar=True)
            record_string += ", sub_pixel_factor=" + str(sub_pixel_factor)
        except TypeError:
            # Hyperspy might not accept 'sub_pixel_factor'
            shifts = s.estimate_shift2D(
                reference=ref,
                roi=(roi.left, roi.right, roi.top, roi.bottom),
                sobel=sobel, hanning=hanning, medfilter=median,
                plot=plot,
                show_progressbar=True)
        s_aligned = signal.deepcopy()
        s_aligned.align2D(shifts=shifts, crop=crop, expand=expand)
        record_string += ", crop={0}, expand={1}".format(crop, expand)
        s_aligned.plot()
        self.record_code("signal = ui.get_selected_signal()")
        self.record_code("s_aligned = signal.deepcopy()")
        self.record_code("s_aligned.align1D(%s)" % record_string)
        return s_aligned

    def align_vertical(self, roi, signal=None):
        signal = self._get_signal(signal)
        if signal is None:
            return
        self.record_code("<p>.align_vertical(roi=%s)" % repr(roi))
        return self._align_along_axis(roi, signal, axis=1)

    def align_horizontal(self, roi, signal=None):
        signal = self._get_signal(signal)
        if signal is None:
            return
        self.record_code("<p>.align_vertical(roi=%s)" % repr(roi))
        return self._align_along_axis(roi, signal, axis=0)

    def _align_along_axis(self, roi, signal, axis):
        sumaxis = 1 if axis == 0 else 0
        daxis = signal.axes_manager.signal_axes[sumaxis]
        iref = daxis.index
        s_al = roi(signal).sum(axis=daxis.index_in_array+3j)
        s_al.change_dtype(float)
        s_al.unfold()   # Temp signal, so don't need to refold
        # Check that signal axis is last dimension
        if s_al.axes_manager.signal_axes[0].index_in_array < 1:
            s_al.data = s_al.data.T             # Unfolded, so simply transpose
        # From now on, navigation is in first dimension
        smooth = self.settings['1d_smooth_amount', float]
        d = np.array([self._smooth(s_al.data[i, :], smooth)
                      for i in range(s_al.data.shape[0])])
        d = np.diff(d, axis=1)      # Differentiate to highlight edges
        sz = d.shape                # Initial shape
        ref = d[iref, :]               # Reference row
        # Pad reference with +/- half size at each ends (maximum shift allowed)
        ref = np.pad(ref, (sz[1] / 2, sz[1] / 2), 'edge')
        shifts = []
        # Find shifts for each row
        for row in range(sz[0]):
            if row == iref:
                # Set shift of reference to compensate for padding
                shifts.append(sz[1] / 2)
            else:
                corr = np.correlate(ref, d[row, :], 'valid')
                shifts.append(corr.argmax())
        # Remove "padding" from found shifts
        shifts = np.array(shifts) - sz[1] / 2
        # Pad for both x and y shifts, but zero unused one:
        shifts = np.tile(-shifts, (2, 1)).T
        shifts[:, axis] = 0.0
        # Apply shifts using hyperspy routine:
        s_aligned = signal.deepcopy()
        s_aligned.align2D(shifts=shifts, crop=False, expand=True)
        s_aligned.plot()
        return s_aligned


class ManualAlignDialog(ExToolWindow):

    def __init__(self, signal, parent=None):
        super(ExToolWindow, self).__init__(parent)
        self.ui = parent
        self.signal = signal
        self._orig_data = None
        self.shifts = None
        self._prev_x = 0
        self._prev_y = 0
        self.create_controls()
        self.accepted.connect(self.ok)
        self.rejected.connect(self.cancel)

    def ok(self):
        """
        Callback when dialog is closed by OK-button.
        """
        signal = self.signal
        if self._orig_data is not None:
            signal.data = self._orig_data
        if self.shifts is not None:
            with signal.unfolded(unfold_signal=False):
                signal.align2D(shifts=self.shifts, expand=True)
            signal.get_dimensions_from_data()
        rc = self.ui.record_code
        rc("signal = ui.get_selected_signal()")
        rc("shifts = np.array(%s)" % str(self.shifts.tolist()))
        rc("with signal.unfolded(unfold_signal=False):")
        rc("    signal.align2D(shifts=shifts, expand=True)")
        rc("signal.get_dimensions_from_data()")

    def cancel(self):
        signal = self.signal
        if self._orig_data is not None:
            signal.data = self._orig_data
            signal.update_plot()
        self.close()

    def close(self):
        self._orig_data = None

    def update_x(self):
        signal = self.signal
        val = self.num_x.value() - self._prev_x
        self._prev_x = self.num_x.value()
        if self._orig_data is None:
            with signal.unfolded(unfold_signal=False):
                self._orig_data = signal.data.copy()
        if self.shifts is None:
            self.shifts = np.zeros((signal.axes_manager.navigation_size, 2),
                                   dtype=np.int)
        index = np.ravel_multi_index(signal.axes_manager.indices,
                                     signal.axes_manager.navigation_shape)
        with signal.unfolded(unfold_signal=False):
            axis = signal.axes_manager.signal_axes[0].index_in_array
            signal.data[index:, ...] = np.roll(
                signal.data[index:, ...], val, axis)
            self.shifts[index:, 1] -= val
            signal.update_plot()

    def update_y(self):
        signal = self.signal
        val = self.num_y.value() - self._prev_y
        self._prev_y = self.num_y.value()
        if self._orig_data is None:
            with signal.unfolded(unfold_signal=False):
                self._orig_data = signal.data.copy()
        if self.shifts is None:
            self.shifts = np.zeros((signal.axes_manager.navigation_size, 2),
                                   dtype=np.int)
        index = np.ravel_multi_index(signal.axes_manager.indices,
                                     signal.axes_manager.navigation_shape)
        with signal.unfolded(unfold_signal=False):
            axis = signal.axes_manager.signal_axes[1].index_in_array
            signal.data[index:, ...] = np.roll(
                signal.data[index:, ...], val, axis)
            self.shifts[index:, 0] -= val
            signal.update_plot()

    def create_controls(self):
        self.setWindowTitle("Align signal")
        form = QtWidgets.QFormLayout()
        self.num_x = QtWidgets.QSpinBox()
        self.num_y = QtWidgets.QSpinBox()
        self.num_x.valueChanged.connect(self.update_x)
        self.num_y.valueChanged.connect(self.update_y)
        dims = self.signal.axes_manager.signal_shape
        self.num_x.setMaximum(dims[0])
        self.num_y.setMaximum(dims[1])
        self.num_x.setMinimum(-dims[0])
        self.num_y.setMinimum(-dims[1])
        form.addRow("X:", self.num_x)
        form.addRow("Y:", self.num_y)
        vbox = QtWidgets.QVBoxLayout()
        vbox.addLayout(form)
        btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                                QtCore.Qt.Horizontal, self)
        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)
        vbox.addWidget(btns)
        self.setLayout(vbox)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Fri Dec 19 03:43:51 2014

@author: Vidar Tonaas Fauske
"""


def _on_figure_window_close(figure, function):
    """Connects a close figure signal to a given function.

    Parameters
    ----------

    figure : mpl figure instance
    function : function

    """
    window = figure.canvas.manager.window
    if not hasattr(figure, '_on_window_close'):
        figure._on_window_close = list()
    if function not in figure._on_window_close:
        figure._on_window_close.append(function)

    window.closing.connect(function)


def override_hyperspy():
    import hyperspy.drawing.utils
    hyperspy.drawing.utils.on_figure_window_close = _on_figure_window_close
    
    from hyperspy.defaults_parser import preferences
    preferences.GUIs.enable_ipywidgets_gui = False
    preferences.GUIs.enable_traitsui_gui = True



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sat Feb 21 17:55:01 2015

@author: Vidar Tonaas Fauske
"""

from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtWidgets import (QDialogButtonBox, QCheckBox, QLineEdit, 
                            QPushButton, QHBoxLayout, QVBoxLayout, QFormLayout,
                            )

import os

from pyqode.core import api
from pyqode.core import modes
from pyqode.core import panels
from pyqode.core.widgets import TabWidget
from . import _editor_server as server
from pyqode.python import modes as pymodes
from pyqode.python.backend.workers import run_pyflakes, calltips

from .extendedqwidgets import ExToolWindow
import hyperspyui.plugincreator as pc


def tr(text):
    return QtCore.QCoreApplication.translate("EditorWidget", text)


class NameCategoryPrompt(ExToolWindow):

    def __init__(self, main_window, parent=None):
        super(NameCategoryPrompt, self).__init__(parent)
        self.setWindowTitle(tr("Plugin properties"))
        self.ui = main_window
        self.create_controls()

    def checks_changed(self):
        menu = self.chk_menu.isChecked()
        toolbar = self.chk_toolbar.isChecked()
        icon = self.chk_icon.isChecked()

        if menu or toolbar:
            self.txt_category.setEnabled(True)
        else:
            self.txt_category.setEnabled(False)

        self.txt_iconpath.setEnabled(icon)
        self.btn_browse_icon.setEnabled(icon)

    def browse_icon(self):
        formats = QtGui.QImageReader.supportedImageFormats()
        formats = [str(f, encoding='ascii') for f in formats]
        # Add one filter that takes all supported:
        type_filter = tr("Supported formats")
        type_filter += ' (*.{0})'.format(' *.'.join(formats))
        # Add all as individual options
        type_filter += ';;' + tr("All files") + \
            ' (*.*) ;;*.' + ';;*.'.join(formats)
        path = QtWidgets.QFileDialog.getOpenFileName(self,
            tr("Pick icon"),
            os.path.dirname(self.ui.cur_dir),
            type_filter)
        if isinstance(path, tuple):    # Pyside returns tuple, PyQt not
            path = path[0]
        if path is None:
            return
        self.txt_iconpath.setText(path)

    def create_controls(self):
        self.txt_name = QLineEdit()
        self.chk_menu = QCheckBox(tr("Menu entry"))
        self.chk_menu.setChecked(True)
        self.chk_toolbar = QCheckBox(tr("Toolbar button"))
        self.chk_toolbar.setChecked(True)
        self.txt_category = QLineEdit()
        self.chk_icon = QCheckBox()
        self.chk_toolbar.setChecked(False)
        self.txt_iconpath = QLineEdit()
        self.btn_browse_icon = QPushButton("...")
        self.txt_iconpath.setEnabled(False)
        self.btn_browse_icon.setEnabled(False)
        btns = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                                QtCore.Qt.Horizontal)

        self.chk_menu.toggled.connect(self.checks_changed)
        self.chk_toolbar.toggled.connect(self.checks_changed)
        self.chk_icon.toggled.connect(self.checks_changed)
        self.btn_browse_icon.clicked.connect(self.browse_icon)
        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)

        hbox = QHBoxLayout()
        for w in [self.chk_icon, self.txt_iconpath, self.btn_browse_icon]:
            hbox.addWidget(w)

        form = QFormLayout()
        form.addRow(tr("Name"), self.txt_name)
        form.addRow(self.chk_menu, self.chk_toolbar)
        form.addRow(tr("Category"), self.txt_category)
        form.addRow(tr("Icon"), hbox)

        vbox = QVBoxLayout(self)
        vbox.addLayout(form)
        vbox.addWidget(btns)

        self.setLayout(vbox)


class ConsoleCodeCheckerMode(modes.CheckerMode):
    """ Runs pyflakes on you code while you're typing

    This checker mode runs pyflakes on the fly to check your python syntax.

    Works the same as FrostedCheckerMode, except it is aware of
    HyperspyUI's "globals".
    """

    def _request(self):
        """ Requests a checking of the editor content. """
        if not self.editor:
            return
        code = self.editor.toPlainText()
        if not self.myeditor.is_plugin:
            code = server._console_mode_header + code
        request_data = {
            'code': code,
            'path': self.editor.file.path,
            'encoding': self.editor.file.encoding
        }
        try:
            self.editor.backend.send_request(
                self._worker, request_data, on_receive=self._on_work_finished)
            self._finished = False
        except NotRunning:
            # retry later
            QtCore.QTimer.singleShot(100, self._request)

    def _on_work_finished(self, results):
        """
        Display results.

        :param status: Response status
        :param results: Response data, messages.
        """
        messages = []
        for msg in results:
            msg = modes.CheckerMessage(*msg)
            msg.line -= server._header_num_lines
            if msg.line < 0:
                continue
            block = self.editor.document().findBlockByNumber(msg.line)
            msg.block = block
            messages.append(msg)
        self.add_messages(messages)

    def __init__(self, editor):
        self.myeditor = editor
        super(ConsoleCodeCheckerMode, self).__init__(run_pyflakes,
                                                     delay=1200)


class ConsoleCodeCalltipsMode(pymodes.CalltipsMode):

    def __init__(self, editor, *args, **kwargs):
        super(ConsoleCodeCalltipsMode, self).__init__(*args, **kwargs)
        self.myeditor = editor

    def _request_calltip(self, source, line, col, fn, encoding):
        if self._CalltipsMode__requestCnt == 0:
            self._CalltipsMode__requestCnt += 1
            if not self.myeditor.is_plugin:
                source = server._console_mode_header + source
                line += server._header_num_lines
            self.editor.backend.send_request(
                calltips,
                {'code': source, 'line': line, 'column': col, 'path': None,
                 'encoding': encoding}, on_receive=self._on_results_available)


class EditorWidget(ExToolWindow):
    code_title = tr("Code Editor")
    plugin_title = tr("Plugin Editor")

    def __init__(self, main_window, parent=None, path=None):
        super(EditorWidget, self).__init__(parent)
        self.setWindowTitle(self.code_title)
        self.ui = main_window
        self._is_plugin = False
        self._suppress_append = False

        self.create_controls(path)

        self.save_action = QtWidgets.QAction(self)
        self.save_action.setShortcut(QtGui.QKeySequence.Save)
        self.save_action.triggered.connect(self.save)
        self.addAction(self.save_action)

        self.run_action = QtWidgets.QAction(self)
        self.run_action.setShortcut(QtGui.QKeySequence(QtCore.Qt.Key_F5))
        self.run_action.triggered.connect(self.run)
        self.addAction(self.run_action)

        if path is not None and os.path.isfile(path):
            self.load(path)

    @property
    def is_plugin(self):
        return self._is_plugin

    @is_plugin.setter
    def is_plugin(self, value):
        if self._is_plugin != value:
            self._is_plugin = value
            self.btn_make_plugin.setVisible(not value)
            self.btn_run.setVisible(not value)
            self.btn_reg_plugin.setVisible(value)
            if value:
                self.setWindowTitle(self.plugin_title)
            else:
                self.setWindowTitle(self.code_title)

    def append_code(self, code):
        if self._suppress_append:
            return
        text = self.editor.toPlainText()
        if len(text) > 0 and text[-1] == '\n':
            prev_cursor = self.editor.textCursor()
            self.editor.moveCursor(QtGui.QTextCursor.End)
            self.editor.insertPlainText(code)
            self.editor.setTextCursor(prev_cursor)
        else:
            self.editor.appendPlainText(code)

    def make_plugin(self):
        diag = NameCategoryPrompt(self.ui, self)
        dr = diag.exec_()
        if dr == QtWidgets.QDialog.Accepted:
            if diag.chk_icon.isChecked():
                icon = diag.txt_iconpath.text()
            else:
                icon = None
            code = "ui = self.ui\n"
            code += "siglist = ui.hspy_signals\n"
            code += self.editor.toPlainText().rstrip()

            name = diag.txt_name.text()
            code = pc.create_plugin_code(code,
                                         name,
                                         diag.txt_category.text(),
                                         diag.chk_menu.isChecked(),
                                         diag.chk_toolbar.isChecked(),
                                         icon)
            path = os.path.normpath(os.path.dirname(__file__) +
                                    '/../plugins/' +
                                    name.lower().replace(' ', '').replace(
                                        '_', '') +
                                    '.py')
            e = EditorWidget(self.ui, self.ui, path)
            e.append_code(code)
            e.editor.moveCursor(QtGui.QTextCursor.Start)
            e.editor.ensureCursorVisible()
            self.ui.editors.append(e)   # We have to keep an instance!
            e.finished.connect(lambda: self.ui.editors.remove(e))
            e.is_plugin = True
            e.show()

    def register_plugin(self):
        dr = self.save()
        path = self.editor.file.path
        if dr and os.path.isfile(path):
            self.ui.plugin_manager.load_from_file(path)

    def save(self):
        if not os.path.isfile(self.editor.file.path):
            path = QtWidgets.QFileDialog.getSaveFileName(self,
                tr("Choose destination path"),
                self.editor.file.path)[0]
            if not path:
                return False
            self.editor.file._path = path
            open(path, 'w').close()  # Touch file so save_current works
        return self.tab.save_current()

    def load(self, filename):
        if filename is not None and os.path.exists(filename):
            self.editor.file.open(filename)

    def run(self):
        code = self.editor.toPlainText()
        old = self._suppress_append
        self._suppress_append = True
        self.ui.console.ex(code)
        self._suppress_append = old

    def sizeHint(self):
        # Default size to fit right margin
        def_sz = super(EditorWidget, self).sizeHint()
        if hasattr(self, 'editor'):
            font = QtGui.QFont(self.editor.font_name, self.editor.font_size +
                               self.editor.zoom_level)
            metrics = QtGui.QFontMetricsF(font)
            pos = 80
            cm = self.layout().contentsMargins()
            # TODO: Currently uses manual, magic number. Do properly!
            offset = self.editor.contentOffset().x() + \
                2 * self.editor.document().documentMargin() + \
                cm.left() + cm.right() + 60
            x80 = round(metrics.width(' ') * pos) + offset
            def_sz.setWidth(x80)
        return def_sz

    def create_controls(self, path):
        editor = api.CodeEdit()
        editor.backend.start(server.__file__)

#        editor.panels.append(panels.FoldingPanel())
        editor.panels.append(panels.LineNumberPanel())
        editor.panels.append(panels.CheckerPanel())

        editor.modes.append(modes.CaretLineHighlighterMode())
        editor.modes.append(modes.CodeCompletionMode())
        editor.modes.append(modes.ExtendedSelectionMode())
        editor.modes.append(modes.FileWatcherMode())
        editor.modes.append(modes.RightMarginMode())
        editor.modes.append(modes.SmartBackSpaceMode())
        editor.modes.append(modes.OccurrencesHighlighterMode())
        editor.modes.append(modes.SymbolMatcherMode())
#        editor.modes.append(modes.WordClickMode())
        editor.modes.append(modes.ZoomMode())

        editor.modes.append(pymodes.PythonSH(editor.document()))
        editor.modes.append(pymodes.CommentsMode())
        editor.modes.append(ConsoleCodeCalltipsMode(self))
        editor.modes.append(ConsoleCodeCheckerMode(self))
        editor.modes.append(pymodes.PEP8CheckerMode())
        editor.modes.append(pymodes.PyAutoCompleteMode())
        editor.modes.append(pymodes.PyAutoIndentMode())
        editor.modes.append(pymodes.PyIndenterMode())

        if path is not None:
            editor.file._path = path
        self.editor = editor
        self.tab = TabWidget(self)
        if path is not None and os.path.isfile(path):
            self.tab.add_code_edit(editor)
        else:
            self.tab.add_code_edit(editor, tr("untitled") + "%d.py")

        self.btn_save = QPushButton(tr("Save"))
        self.btn_run = QPushButton(tr("Run"))
        self.btn_make_plugin = QPushButton(tr("Make Plugin"))
        self.btn_reg_plugin = QPushButton(tr("Register Plugin"))
        self.btn_reg_plugin.setVisible(False)

        self.btn_save.clicked.connect(self.save)
        self.btn_run.clicked.connect(self.run)
        self.btn_make_plugin.clicked.connect(self.make_plugin)
        self.btn_reg_plugin.clicked.connect(self.register_plugin)

        self.hbox = QHBoxLayout()
        for w in [self.btn_save, self.btn_run,
                  self.btn_make_plugin, self.btn_reg_plugin]:
            self.hbox.addWidget(w)

        vbox = QVBoxLayout(self)
        vbox.addWidget(self.tab)
        vbox.addLayout(self.hbox)

        self.setLayout(vbox)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Wed Jan 20 02:43:41 2016

@author: Vidar Tonaas Fauske
"""

from qtpy import QtCore, QtWidgets
from qtpy.QtWidgets import QDialogButtonBox


class StringInputDialog(QtWidgets.QDialog):

    def __init__(self, prompt="", default="", parent=None):
        super(StringInputDialog, self).__init__(parent=parent)
        self.setWindowTitle("Input prompt")
        self.setWindowFlags(QtCore.Qt.Tool)

        frm = QtWidgets.QFormLayout()
        self.edit = QtWidgets.QLineEdit(default)
        frm.addRow(prompt, self.edit)

        btns = QDialogButtonBox(
                QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                QtCore.Qt.Horizontal, self)
        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)

        box = QtWidgets.QVBoxLayout(self)
        box.addLayout(frm)
        box.addWidget(btns)
        self.setLayout(box)

    def prompt_modal(self, rejection=None):
        dr = self.exec_()
        if dr == QtWidgets.QDialog.Accepted:
            return self.edit.text()
        else:
            return rejection

    def _on_completed(self, result):
        (callback, rejection) = self._on_completed_info
        if result == QtWidgets.QDialog.Accepted:
            value = self.edit.text()
        else:
            value = rejection
        callback(value)

    def prompt_modeless(self, callback, rejection=None):
        self._on_completed_info = (callback, rejection)
        dr = self.show()
        dr.finished.connect(self._on_completed)



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Mon May 04 17:30:36 2015

@author: Vidar Tonaas Fauske
"""

from hyperspyui.plugins.plugin import Plugin

from qtpy import QtCore, QtNetwork, QtWidgets
from qtpy.QtWebEngineWidgets import QWebEnginePage, QWebEngineView, WEBENGINE


try:
    assert(QtNetwork.QSslSocket.supportsSsl())
except NameError:
    raise ImportError("Current platform doesn't support SSL. EELSDB plugin"
                      " disabled.")

from hyperspyui.widgets.extendedqwidgets import ExToolWindow
import os
import re
import urllib.request, urllib.error, urllib.parse
import tempfile

re_dl_url = re.compile(
    br'https?://eelsdb\.eu/wp-content/uploads/\d{4}/\d{2}/.+\.msa')


class EELSDBPlugin(Plugin):
    name = "EELSDB Plugin"

    def __init__(self, main_window):
        super(EELSDBPlugin, self).__init__(main_window)
        self.window = None
        self.view = None

    def create_actions(self):
        self.add_action(self.name + '.browse', "Browse EELSDB", self.default,
                        tip="Browse the EELSDB online database of standard"
                        "EEL spectra")

    def create_menu(self):
        self.add_menuitem('EELS', self.ui.actions[self.name + '.browse'])

    def _make_request(self, url):
        request = QtNetwork.QNetworkRequest()
        request.setUrl(QtCore.QUrl(url))
        return request

    def load_blocking(self, view, *args):
        loop = QtCore.QEventLoop()
        view.loadFinished.connect(loop.quit)

        def unravel():
            view.loadFinished.disconnect(loop.quit)
            view.loadFinished.disconnect(unravel)

        view.loadFinished.connect(unravel)
        view.load(*args)
        loop.exec_()

    def _on_link(self, url):
        su = url.toString(QtCore.QUrl.RemoveQuery)
        if "/spectra/" in su:
            view = QWebEngineView(self.ui)
            wp = QWebEnginePage()
            am = self.view.page().networkAccessManager()
            wp.setNetworkAccessManager(am)
            view.setPage(wp)
            request = self._make_request(url)
            response = am.get(request)

            def resp_finished():
                html = response.readAll()
                matches = re_dl_url.findall(html)
                if matches:
                    self.download(matches[0].decode('ascii'))
                response.deleteLater()
            response.finished.connect(resp_finished)
        else:
            self.view.load(url)

    def download(self, url):
        header = {'Accept': r'text/html,application/xhtml+xml,'
                  'application/xml;q=0.9,*/*;q=0.8',
                  'Accept-Encoding': r'gzip, deflate',
                  'User-Agent': r'Mozilla/5.0 (Windows NT 10.0; WOW64;'
                  ' rv:43.0)Gecko/20100101 Firefox/43.0',
                  }

        suffix = '.msa'

        req = urllib.request.Request(url, headers=header)
        page = urllib.request.urlopen(req)

        try:
            f = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
            fn = f.name
            f.write(page.read())
            fn = f.name
            f.close()
            self.ui.load([fn])
        finally:
            f.close()
            os.remove(fn)
        self.record_code("<p>.download(url='%s')" % url)

    def default(self):
        if self.window is None:
            self.window = ExToolWindow(self.ui)
            self.window.setWindowTitle("EELSDB")
            vbox = QtWidgets.QVBoxLayout()
            self.window.setLayout(vbox)
            view = QWebEngineView(self.ui)
            self.view = view
            vbox.addWidget(view)
            self.window.resize(self.view.sizeHint())
        # Load spectra browser
        browse_url = QtCore.QUrl("https://eelsdb.eu/spectra")
        self.load_blocking(self.view, browse_url)
        if not WEBENGINE:
            self.view.page().setLinkDelegationPolicy(
                    QWebEnginePage.DelegateAllLinks)
        try:
            # TODO: downloading spectra is currently broken
            self.view.linkClicked.connect(self._on_link)
        except AttributeError:
            pass
        self.window.show()



# -*- coding: utf-8 -*-
# Copyright 2014-2016 The HyperSpyUI developers
#
# This file is part of HyperSpyUI.
#
# HyperSpyUI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpyUI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpyUI.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Tue Nov 04 16:25:54 2014

@author: Vidar Tonaas Fauske
"""


from qtpy import QtCore
#from hyperspy.model import Model
import hyperspy.models.eelsmodel
from .actionable import Actionable
from functools import partial

from hyperspyui.widgets.stringinput import StringInputDialog

# TODO: Add smartfit for EELSModel


def tr(text):
    return QtCore.QCoreApplication.translate("ModelWrapper", text)


class ModelWrapper(Actionable):
    added = QtCore.Signal((object, object), (object,))
    removed = QtCore.Signal((object, object), (object,))

    def __init__(self, model, signal_wrapper, name):
        super(ModelWrapper, self).__init__()
        self.model = model
        self.signal = signal_wrapper
        self.name = name
        if self.signal.signal is not self.model.signal:
            raise ValueError("SignalWrapper doesn't match model.signal")
        self.components = {}
        self.update_components()

        self.fine_structure_enabled = False

        # Default actions
        self.add_action('plot', tr("&Plot"), self.plot)
        self.add_action('fit', tr("&Fit"), self.fit)
        self.add_action('multifit', tr("&Multifit"), self.multifit)
        self.add_action('set_signal_range', tr("Set signal &range"),
                        self.set_signal_range)
        if isinstance(self.model, hyperspy.models.eelsmodel.EELSModel):
            self.add_action('lowloss', tr("Set low-loss"), self.set_lowloss)
            self.add_action('fine_structure', tr("Enable fine &structure"),
                            self.toggle_fine_structure)
        f = partial(self.signal.remove_model, self)
        self.add_action('delete', tr("&Delete"), f)

    def plot(self):
        self.signal.keep_on_close = True
        self.model.plot()
        self.signal.keep_on_close = False
        self.signal.update_figures()
        self.signal.signal_plot.setProperty('hyperspyUI.ModelWrapper', self)

    def update_plot(self):
        self.model.update_plot()

    def record_code(self, code):
        self.signal.mainwindow.record_code("model = ui.get_selected_model()")
        self.signal.mainwindow.record_code(code)

    def _args_for_record(self, args, kwargs):
        argstr = str(args)[1:-1]
        kwargstr = str(kwargs)[1:-1]
        kwargstr = kwargstr.replace(": ", "=")
        if argstr and kwargstr:
            return ", ".join((argstr, kwargstr))
        else:
            return argstr + kwargstr

    def fit(self, *args, **kwargs):
        self.signal.keep_on_close = True
        self.model.fit(*args, **kwargs)
        self.signal.keep_on_close = False
        self.signal.update_figures()
        self.record_code("model.fit(%s)" %
                         self._args_for_record(args, kwargs))

    def multifit(self, *args, **kwargs):
        self.signal.keep_on_close = True
        self.model.multifit(*args, **kwargs)
        self.signal.keep_on_close = False
        self.signal.update_figures()
        self.record_code("model.multifit(%s)" %
                         self._args_for_record(args, kwargs))

    def smartfit(self, *args, **kwargs):
        if hasattr(self.model, 'smartfit'):
            self.signal.keep_on_close = True
            self.model.smartfit(*args, **kwargs)
            self.signal.keep_on_close = False
            self.signal.update_figures()
            self.record_code("model.smartfit(%)" %
                             self._args_for_record(args, kwargs))

    def fit_component(self, component):
        # This is a non-blocking call, which means the normal keep_on_close +
        # update_figures won't work. To make sure we keep our figures,
        # we force a plot first if it is not active already.
        if not self.model.signal._plot.is_active:
            self.plot()
        self.model.fit_component(component)
        self.record_code("model.fit_component(%s)" % component.name)

    def set_signal_range(self, *args, **kwargs):
        self.signal.keep_on_close = True
        self.model.set_signal_range(*args, **kwargs)
        self.signal.keep_on_close = False
        self.signal.update_figures()
        self.record_code("model.set_signal_range(%s)" %
                         self._args_for_record(args, kwargs))

    def set_lowloss(self, signal=None):
        if signal is None:
            signal = self.signal.mainwindow.select_x_signals(
                1, ['Select low-loss'])
            if signal is None:
                return
        self.model.lowloss = signal.signal
        self.record_code("model.set_lowloss(low_loss_signal)")

    def toggle_fine_structure(self):
        if not isinstance(self.model, hyperspy.models.eelsmodel.EELSModel):
            raise TypeError(
                tr("Model is not EELS model. Can not toggle fine structure"))
        if self.fine_structure_enabled:
            self.model.disable_fine_structure()
            self.actions['fine_structure'].setText(
                tr("Enable fine &structure"))
            self.record_code("model.disable_fine_structure()")
        else:
            self.model.enable_fine_structure()
            self.actions['fine_structure'].setText(
                tr("Disable fine &structure"))
            self.record_code("model.enable_fine_structure()")
        self.fine_structure_enabled = not self.fine_structure_enabled

    def update_components(self):
        """
        Updates internal compoenent list to match model's list (called e.g.
        after console execute and in constructor)
        """

        # Add missing
        for c in self.model:
            if c.name not in list(self.components.keys()):
                self.components[c.name] = c
                self.component_added(c)

        # Remove lingering
        ml = [c.name for c in self.model]
        rm = [cn for cn in self.components.keys() if cn not in ml]
        for n in rm:
            c = self.components.pop(n)
            self.component_removed(c)

    def add_component(self, component):
        if isinstance(component, type):
            nec = ['EELSCLEdge', 'Spline', 'ScalableFixedPattern']
            if component.__name__ in nec:
                raise TypeError(
                    tr("Component of type %s currently not supported")
                    % component)
            elif component.__name__ == 'Expression':
                dlg = StringInputDialog(prompt="Enter expression:")
                expression = dlg.prompt_modal(rejection=None)
                if expression:
                    component = component(expression, 'Expression')
                else:
                    return
            else:
                component = component()

        added = False
        if component not in self.model:
            self.model.append(component)
            added = True
            self.record_code("model.append(%s)" % component.name)
        if component.name not in self.components:
            self.components[component.name] = component
            added = True
        if added:
            self.component_added(component)

    def remove_component(self, component):
        removed = False
        if component in self.model:
            self.model.remove(component)
            self.record_code("model.remove(%s)" % component.name)
            removed = True
        if component.name in self.components:
            self.components.pop(component.name)
            removed = True
        if removed:
            self.component_removed(component)

    def component_added(self, component):
        self.update_plot()
        self.added[object, object].emit(component, self)
        self.added[object].emit(component)

    def component_removed(self, component):
        self.update_plot()
        self.removed[object, object].emit(component, self)
        self.removed[object].emit(component)
